# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
HTTP server.
"""

import os
import socket
import threading
import struct
import logging

from .base import Session, Request, _handle_requests, _TYPE_ERROR


__all__ = ('Server', 'SSLServer', 'Session', 'Request')
log = logging.getLogger(__name__)


def build_server_sslctx(sslconfig):
    """
    Build an `ssl.SSLContext` appropriately configured for server-side use.

    For example:

    >>> sslconfig = {
    ...     'cert_file': '/my/server.cert',
    ...     'key_file': '/my/server.key',
    ...     'ca_file': '/my/client.ca',
    ... }
    >>> sslctx = build_server_sslctx(sslconfig)  #doctest: +SKIP

    """
    # Lazily import `ssl` module to be memory friendly when SSL isn't needed:
    import ssl

    if not isinstance(sslconfig, dict):
        raise TypeError(
            _TYPE_ERROR.format('sslconfig', dict, type(sslconfig), sslconfig)
        )

    # For safety and clarity, force all paths to be absolute, normalized paths:
    for key in ('cert_file', 'key_file', 'ca_file', 'ca_path'):
        if key in sslconfig:
            value = sslconfig[key]
            if value != os.path.abspath(value):
                raise ValueError(
                    'sslconfig[{!r}] is not an absulute, normalized path: {!r}'.format(
                        key, value
                    )
                )

    sslctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
    sslctx.set_ciphers(
        'ECDHE-RSA-AES128-GCM-SHA256:ECDHE-RSA-AES256-GCM-SHA384'
    )
    sslctx.set_ecdh_curve('secp384r1')
    sslctx.options |= ssl.OP_NO_COMPRESSION
    sslctx.options |= ssl.OP_SINGLE_ECDH_USE
    sslctx.options |= ssl.OP_CIPHER_SERVER_PREFERENCE
    sslctx.load_cert_chain(sslconfig['cert_file'], sslconfig['key_file'])
    if 'allow_unauthenticated_clients' in sslconfig:
        if sslconfig['allow_unauthenticated_clients'] is not True:
            raise ValueError(
                'True is only allowed value for allow_unauthenticated_clients'
            )
        if {'ca_file', 'ca_path'}.intersection(sslconfig):
            raise ValueError(
                'ca_file/ca_path with allow_unauthenticated_clients'
            )
        return sslctx
    if not {'ca_file', 'ca_path'}.intersection(sslconfig):
        raise ValueError(
            'must include ca_file or ca_path (or allow_unauthenticated_clients)'
        )
    sslctx.verify_mode = ssl.CERT_REQUIRED
    sslctx.load_verify_locations(
        cafile=sslconfig.get('ca_file'),
        capath=sslconfig.get('ca_path'),
    )
    return sslctx


def _validate_server_sslctx(sslctx):
    # Lazily import `ssl` module to be memory friendly when SSL isn't needed:
    import ssl

    if isinstance(sslctx, dict):
        sslctx = build_server_sslctx(sslctx)

    if not isinstance(sslctx, ssl.SSLContext):
        raise TypeError('sslctx must be an ssl.SSLContext')
    if sslctx.protocol != ssl.PROTOCOL_TLSv1_2:
        raise ValueError('sslctx.protocol must be ssl.PROTOCOL_TLSv1_2')

    # We consider ssl.CERT_OPTIONAL to be a bad grey area:
    if sslctx.verify_mode == ssl.CERT_OPTIONAL:
        raise ValueError('sslctx.verify_mode cannot be ssl.CERT_OPTIONAL')
    assert sslctx.verify_mode in (ssl.CERT_REQUIRED, ssl.CERT_NONE)
    if sslctx.verify_mode != ssl.CERT_REQUIRED:
        log.warning('Security concern: sslctx allows unauthenticated clients!')

    # Check the options:
    if not (sslctx.options & ssl.OP_NO_COMPRESSION):
        raise ValueError('sslctx.options must include ssl.OP_NO_COMPRESSION')
    if not (sslctx.options & ssl.OP_SINGLE_ECDH_USE):
        raise ValueError('sslctx.options must include ssl.OP_SINGLE_ECDH_USE')
    if not (sslctx.options & ssl.OP_CIPHER_SERVER_PREFERENCE):
        raise ValueError('sslctx.options must include ssl.OP_CIPHER_SERVER_PREFERENCE')

    return sslctx


def _fq_name(app):
    if hasattr(app, '__qualname__'):
        return '{}.{}'.format(app.__module__, app.__qualname__)
    cls = app.__class__
    return '{}.{}(<...>)'.format(cls.__module__, cls.__qualname__)


def _get_credentials(sock):
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_PASSCRED, 1)
    size = struct.calcsize('3i')
    data = sock.getsockopt(socket.SOL_SOCKET, socket.SO_PEERCRED, size)
    return struct.unpack('3i', data)


def _shutdown_and_close(sock):
    try:
        sock.shutdown(socket.SHUT_RDWR)
    except OSError:
        pass
    sock.close()


class Server:
    _options = ('max_connections', 'max_requests', 'timeout')
    __slots__ = ('address', 'app', 'options', 'sock') + _options

    def __init__(self, address, app, **options):
        # address:
        if isinstance(address, tuple):  
            if len(address) == 4:
                family = socket.AF_INET6
            elif len(address) == 2:
                family = socket.AF_INET
            else:
                raise ValueError(
                    'address: must have 2 or 4 items; got {!r}'.format(address)
                )
        elif isinstance(address, str):
            if os.path.abspath(address) != address:
                raise ValueError(
                    'address: bad socket filename: {!r}'.format(address)
                )
            family = socket.AF_UNIX
        elif isinstance(address, bytes):
            family = socket.AF_UNIX
        else:
            raise TypeError(
                _TYPE_ERROR.format('address', (tuple, str, bytes), type(address), address)
            )

        # app:
        if not callable(app):
            raise TypeError('app: not callable: {!r}'.format(app))
        on_connect = getattr(app, 'on_connect', None)
        if not (on_connect is None or callable(on_connect)):
            raise TypeError('app.on_connect: not callable: {!r}'.format(app))
        self.app = app

        # options:
        if not set(options).issubset(self.__class__._options):
            unsupported = sorted(set(options) - set(self.__class__._options))
            raise TypeError(
                'unsupported {}() **options: {}'.format(
                    self.__class__.__name__, ', '.join(unsupported)
                )
            )
        self.options = options
        self.max_connections = options.get('max_connections', 50)
        self.max_requests = options.get('max_requests', 1000)
        self.timeout = options.get('timeout', 30)
        assert isinstance(self.max_connections, int) and self.max_connections > 0
        assert isinstance(self.max_requests, int) and self.max_requests > 0 
        assert isinstance(self.timeout, (int, float)) and self.timeout > 0

        # Listen...
        self.sock = socket.socket(family, socket.SOCK_STREAM)
        self.sock.bind(address)
        self.address = self.sock.getsockname()
        self.sock.listen(5)

    def __repr__(self):
        return '{}({!r}, {})'.format(
            self.__class__.__name__, self.address, _fq_name(self.app)
        )

    def serve_forever(self):
        try:
            self._serve_forever()
        finally:
            _shutdown_and_close(self.sock)

    def _serve_forever(self):
        log.info('Starting Degu %s @ %r', self.__class__.__name__, self.address)
        log.info('[timeout=%r, max_connections=%r, max_requests=%r]',
            self.timeout, self.max_connections, self.max_requests
        )
        listensock = self.sock
        unix = (True if listensock.family == socket.AF_UNIX else False)
        semaphore = threading.BoundedSemaphore(self.max_connections)
        timeout = self.timeout
        max_requests = self.max_requests
        worker = self._worker
        while True:
            (sock, address) = listensock.accept()
            # Denial of Service note: when we already have max_connections, we
            # should aggressively rate-limit the handling of new connections, so
            # that's why we use `timeout=2` rather than `blocking=False`:
            if semaphore.acquire(timeout=2) is True:
                sock.settimeout(timeout)
                credentials = (_get_credentials(sock) if unix is True else None)
                session = Session(address, credentials, max_requests)
                thread = threading.Thread(
                    target=worker,
                    args=(semaphore, session, sock),
                    daemon=True,
                )
                thread.start()
            else:
                log.warning('Too many connections, rejecting %r', address)
                _shutdown_and_close(sock)

    def _serve_one(self):
        unix = (True if self.sock.family == socket.AF_UNIX else False)
        (sock, address) = self.sock.accept()
        sock.settimeout(self.timeout)
        credentials = (_get_credentials(sock) if unix is True else None)
        session = Session(address, credentials, self.max_requests)
        self._worker(None, session, sock)

    def _worker(self, semaphore, session, sock):
        try:
            log.info('+ %s New connection', session)
            self._handle_connection(session, sock)
        except (socket.timeout, ConnectionError) as e:
            log.info('- %s Handled %d requests: %r',
                session, session.requests, e
            )
        except:
            log.exception('- %s Error after handling %d requests:',
                session, session.requests
            )
        finally:
            _shutdown_and_close(sock)
            if semaphore is not None:
                semaphore.release()

    def _handle_connection(self, session, sock):
        on_connect = getattr(self.app, 'on_connect', None)
        if on_connect is None or on_connect(session, sock) is True:
            _handle_requests(self.app, session, sock)
            sock.close()
            log.info('- %s Handled %d requests: %s',
                session, session.requests, session.message
            )
        else:
            log.warning('- %s Rejected by app.on_connect()', session)


class SSLServer(Server):
    __slots__ = ('sslctx',)

    def __init__(self, sslctx, address, app, **options):
        self.sslctx = _validate_server_sslctx(sslctx)
        super().__init__(address, app, **options)

    def __repr__(self):
        return '{}(<sslctx>, {!r}, {})'.format(
            self.__class__.__name__, self.address, _fq_name(self.app)
        )

    def _handle_connection(self, session, sock):
        sock = self.sslctx.wrap_socket(sock, server_side=True)
        super()._handle_connection(session, sock)




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Pure-Python equivalent of the `degu._base` C extension.

Although `degu._basepy` is automatically imported as a fall-back when the
`degu._base` C extension isn't available, this Python implementation really
isn't meant for production use (mainly because it's much, much slower).

This is a reference implementation whose purpose is only to help enforce the
correctness of the C implementation.
"""

from collections import namedtuple
import socket


TYPE_ERROR = '{}: need a {!r}; got a {!r}: {!r}'
TYPE_ERROR2 = '{}: need a {!r}; got a {!r}'

BUF_LEN = 32768  # 32 KiB
SCRATCH_LEN = 32
MAX_LINE_LEN = 4096  # Max length of chunk size line, including CRLF

MAX_CL_LEN = 16  # Max length (in bytes) of a content-length/etc
MAX_LENGTH = 9999999999999999  # Max value for content-length/etc

MAX_HEADER_COUNT = 20

IO_SIZE = 1048576  # 1 MiB
MAX_IO_SIZE = 16777216  # 16 MiB

BODY_READY = 0
BODY_STARTED = 1
BODY_CONSUMED = 2
BODY_ERROR = 3

_METHODS = {
    b'GET': 'GET',
    b'PUT': 'PUT',
    b'POST': 'POST',
    b'HEAD': 'HEAD',
    b'DELETE': 'DELETE',
}

_OK = 'OK'


BodiesType = Bodies = namedtuple('Bodies',
    'Body ChunkedBody BodyIter ChunkedBodyIter'
)
ResponseType = Response = namedtuple('Response', 'status reason headers body')


class EmptyPreambleError(ConnectionError):
    pass


################    BEGIN GENERATED TABLES    ##################################
NAME = frozenset(
    b'-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
)

DECIMAL = frozenset(b'0123456789')
HEXADECIMAL = frozenset(b'0123456789ABCDEFabcdef')

_LOWER = b'-0123456789abcdefghijklmnopqrstuvwxyz'
_UPPER = b'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
_URI   = b'/?'
_PATH  = b'+.:_~'
_QUERY = b'%&='
_SPACE = b' '
_VALUE = b'"\'()*,;<>[]'

KEY    = frozenset(_LOWER)
VAL    = frozenset(_LOWER + _UPPER + _PATH + _QUERY + _URI + _SPACE + _VALUE)
URI    = frozenset(_LOWER + _UPPER + _PATH + _QUERY + _URI)
PATH   = frozenset(_LOWER + _UPPER + _PATH)
QUERY  = frozenset(_LOWER + _UPPER + _PATH + _QUERY)
REASON = frozenset(_LOWER + _UPPER + _SPACE)
EXTKEY = frozenset(_LOWER + _UPPER)
EXTVAL = frozenset(_LOWER + _UPPER + _PATH + _VALUE)
################    END GENERATED TABLES      ##################################


################################################################################
# Argument type and value checking:

def _check_type(name, obj, _type):
    if type(obj) is not _type:
        raise TypeError(
            TYPE_ERROR.format(name, _type, type(obj), obj)
        )

def _check_type2(name, obj, _type):
    if type(obj) is not _type:
        raise TypeError(
            TYPE_ERROR2.format(name, _type, type(obj))
        )

def _check_tuple(name, obj, size):
    _check_type2(name, obj, tuple)
    if len(obj) != size:
        raise ValueError(
            '{}: need a {}-tuple; got a {}-tuple'.format(name, size, len(obj))
        )
    return obj

def _check_list(name, obj):
    _check_type2(name, obj, list)
    return obj

def _check_bytes(name, obj, max_len=MAX_IO_SIZE):
    assert max_len <= MAX_IO_SIZE
    _check_type2(name, obj, bytes)
    if len(obj) > max_len:
        raise ValueError(
            'need len({}) <= {}; got {}'.format(name, max_len, len(obj))
        )

_OUTGOING_STR = frozenset(chr(i) for i in range(128))

def _check_str(name, obj):
    _check_type(name, obj, str)
    if obj == '' or not _OUTGOING_STR.issuperset(obj):
        raise ValueError(
            'bad {}: {!r}'.format(name, obj)
        )
    return obj

def _getcallable(objname, obj, name):
    attr = getattr(obj, name)
    if not callable(attr):
        raise TypeError('{}.{}() is not callable'.format(objname, name))
    return attr


################################################################################
# Header parsing:

def _validate_int(name, obj):
    if type(obj) is not int:
        raise TypeError(
            TYPE_ERROR.format(name, int, type(obj), obj)
        )

def _validate_length(name, length):
    _validate_int(name, length)
    if not (0 <= length <= MAX_LENGTH):
        raise ValueError(
            'need 0 <= {} <= {}; got {}'.format(name, MAX_LENGTH, length)
        )
    return length

def _validate_size(name, size, max_size):
    assert 0 <= max_size <= MAX_IO_SIZE
    _validate_int(name, size)
    if not (0 <= size <= max_size):
        raise ValueError(
            'need 0 <= {} <= {}; got {}'.format(name, max_size, size)
        )
    return size

def _validate_exact_size(name, size, expected):
    assert 0 <= expected <= MAX_IO_SIZE
    _validate_int(name, size)
    if size != expected:
        raise ValueError(
            'need {} == {!r}; got {!r}'.format(name, expected, size)
        )
    return size

def _validate_read_size(name, size, remaining):
    if size is None:
        if remaining > MAX_IO_SIZE:
            raise ValueError(
                'would exceed max read size: {} > {}'.format(
                    remaining, MAX_IO_SIZE
                )
            )
        return remaining
    return _validate_size(name, size, MAX_IO_SIZE)


def _recv_into(method, dst):
    max_size = len(dst)
    size = method(dst)
    return _validate_size('received', size, max_size)


def _readinto(method, dst):
    dst = memoryview(dst)
    start = 0
    stop = len(dst)
    while start < stop:
        received = _recv_into(method, dst[start:])
        if received == 0:
            break
        start += received
    if start != stop:
        raise ValueError(
            'expected to read {} bytes, but received {}'.format(stop, start)
        )
    return start

def _readinto_from(robj, dst):
    if type(robj) is Reader:
        return robj.readinto(dst)
    return _readinto(robj, dst)


def _send(method, src):
    assert len(src) > 0, src
    max_size = len(src)
    size = method(src)
    return _validate_size('sent', size, max_size)


def _write(method, src):
    src = memoryview(src)
    start = 0
    stop = len(src)
    while start < stop:
        sent = _send(method, src[start:])
        if sent == 0:
            break
        start += sent
    if start != stop:
        raise ValueError(
            'expected to write {} bytes, but sent {}'.format(stop, start)
        )
    return start


def _write_to(wobj, src):
    if len(src) == 0:
        return 0
    if type(wobj) is Writer:
        return wobj._write(src)
    return _write(wobj, src)


def _get_robj(rfile):
    if type(rfile) is Reader:
        return rfile
    return _getcallable('rfile', rfile, 'readinto')


def _get_readline(rfile):
    if type(rfile) is Reader:
        return rfile
    return _getcallable('rfile', rfile, 'readline')


def _get_wobj(wfile):
    if type(wfile) is Writer:
        return wfile
    return _getcallable('wfile', wfile, 'write')


class Range:
    __slots__ = ('_start', '_stop')

    def __init__(self, start, stop):
        start = _validate_length('start', start)
        stop = _validate_length('stop', stop)
        if start >= stop:
            raise ValueError(
                'need start < stop; got {} >= {}'.format(start, stop)
            )
        self._start = start
        self._stop = stop

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    def __repr__(self):
        return 'Range({}, {})'.format(self._start, self._stop)

    def __str__(self):
        return 'bytes={}-{}'.format(self._start, self._stop - 1)

    def __eq__(self, other):
        if type(other) is type(self):
            return self._start == other._start and self._stop == other._stop
        if type(other) is str:
            return str(self) == other
        raise TypeError('cannot compare Range() with {!r}'.format(type(other)))

    def __ne__(self, other):
        return not self.__eq__(other)


class ContentRange:
    __slots__ = ('_start', '_stop', '_total')

    def __init__(self, start, stop, total):
        _validate_length('start', start)
        _validate_length('stop', stop)
        _validate_length('total', total)
        if not (start < stop <= total):
            raise ValueError(
                'need start < stop <= total; got ({}, {}, {})'.format(
                    start, stop, total
                )
            )
        self._start = start
        self._stop = stop
        self._total = total

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @property
    def total(self):
        return self._total

    def __repr__(self):
        return 'ContentRange({}, {}, {})'.format(
            self._start, self._stop, self._total
        )

    def __str__(self):
        return 'bytes {}-{}/{}'.format(
            self._start, self._stop - 1, self._total
        )

    def __eq__(self, other):
        if type(other) is type(self):
            return (self._start == other._start
                and self._stop == other._stop
                and self._total == other._total
            )
        if type(other) is str:
            return str(self) == other
        raise TypeError(
            'cannot compare ContentRange() with {!r}'.format(type(other))
        )

    def __ne__(self, other):
        return not self.__eq__(other)


def _parse_key(src):
    if len(src) < 1:
        raise ValueError('header name is empty')
    if len(src) > 32:
        raise ValueError('header name too long: {!r}...'.format(src[:32]))
    if NAME.issuperset(src):
        return src.decode('ascii').lower()
    raise ValueError('bad bytes in header name: {!r}'.format(src))


def parse_header_name(src):
    """
    Used to decode, validate, and case-fold header keys.

    FIXME: drop from public API, replaced by _parse_key().
    """
    return _parse_key(src)


def _parse_val(src):
    if len(src) < 1:
        raise ValueError('header value is empty')
    if VAL.issuperset(src):
        return src.decode('ascii')
    raise ValueError('bad bytes in header value: {!r}'.format(src))


def parse_chunk_size(src):
    L = len(src)
    if (L > 7):
        raise ValueError('chunk_size is too long: {!r}...'.format(src[:7]))
    if not (HEXADECIMAL.issuperset(src) and L >= 1 and (src[0] != 48 or L == 1)):
        raise ValueError('bad chunk_size: {!r}'.format(src))
    size = int(src, 16)
    if size > MAX_IO_SIZE:
        raise ValueError(
            'need chunk_size <= {}; got {}'.format(MAX_IO_SIZE, size)
        )
    assert size >= 0
    return size


def _parse_chunk_extension_key(src):
    if EXTKEY.issuperset(src):
        return src.decode()
    raise ValueError('bad chunk extension key: {!r}'.format(src))


def _parse_chunk_extension_val(src):
    if EXTVAL.issuperset(src):
        return src.decode()
    raise ValueError('bad chunk extension value: {!r}'.format(src))


def parse_chunk_extension(src):
    assert type(src) is bytes
    parts = src.split(b'=', 1)
    if len(parts) == 2 and parts[0] and parts[1]:
        key = _parse_chunk_extension_key(parts[0])
        val = _parse_chunk_extension_val(parts[1])
        return (key, val)
    raise ValueError('bad chunk extension: {!r}'.format(src))


def parse_chunk(src):
    assert type(src) is bytes
    if len(src) < 1:
        raise ValueError('{!r} not found in {!r}...'.format(b'\r\n', b''))
    parts = src.split(b';', 1)
    size = parse_chunk_size(parts[0])
    if len(parts) == 2:
        ext = parse_chunk_extension(parts[1])
    else:
        ext = None
    return (size, ext)


def _parse_decimal(src):
    if len(src) < 1 or len(src) > 16:
        return -1
    if not DECIMAL.issuperset(src):
        return -1
    if len(src) > 1 and src[0:1] == b'0':
        return -1
    return int(src)


def parse_content_length(src):
    assert isinstance(src, bytes)
    if len(src) > 16:
        raise ValueError(
            'content-length too long: {!r}...'.format(src[:16])
        )
    value = _parse_decimal(src)
    if value < 0:
        raise ValueError('bad content-length: {!r}'.format(src))
    return value


def _raise_bad_range(src):
    raise ValueError('bad range: {!r}'.format(src))


def parse_range(src):
    assert isinstance(src, bytes)
    if len(src) > 39:
        raise ValueError('range too long: {!r}...'.format(src[:39]))
    if len(src) < 9 or src[0:6] != b'bytes=':
        _raise_bad_range(src)
    inner = src[6:]
    parts = inner.split(b'-', 1)
    if len(parts) != 2:
        _raise_bad_range(src)
    start = _parse_decimal(parts[0])
    end = _parse_decimal(parts[1])
    if start < 0 or end < start or end >= MAX_LENGTH:
        _raise_bad_range(src)
    return Range(start, end + 1)


def _bad_content_range(src):
    raise ValueError('bad content-range: {!r}'.format(src))

def parse_content_range(src):
    assert isinstance(src, bytes)
    if len(src) > 56:
        raise  ValueError('content-range too long: {!r}...'.format(src[:56]))
    if len(src) < 11 or src[0:6] != b'bytes ':
        _bad_content_range(src)
    inner = src[6:]
    a = inner.split(b'-', 1)
    if len(a) != 2:
        _bad_content_range(src)
    b = a[1].split(b'/', 1)
    if len(b) != 2:
        _bad_content_range(src)
    start = _parse_decimal(a[0])
    end = _parse_decimal(b[0])
    total = _parse_decimal(b[1])
    if start < 0 or end < start or end >= total or total > MAX_LENGTH:
        _bad_content_range(src)
    return ContentRange(start, end + 1, total)


def _parse_header_lines(header_lines, isresponse=False):
    if type(isresponse) is not bool:
        raise TypeError(
            TYPE_ERROR.format('isresponse', bool, type(isresponse), isresponse)
        )
    headers = {}
    flags = 0
    for line in header_lines:
        if len(line) < 4:
            raise ValueError('header line too short: {!r}'.format(line))
        parts = line.split(b': ', 1)
        if len(parts) != 2:
            raise ValueError('bad header line: {!r}'.format(line))
        (key, val) = parts
        key = _parse_key(key)
        if key == 'content-length':
            flags |= 1
            val = parse_content_length(val)
        elif key == 'transfer-encoding':
            flags |= 2
            if val != b'chunked':
                raise ValueError(
                    'bad transfer-encoding: {!r}'.format(val)
                )
            val = 'chunked'
        elif key == 'range':
            flags |= 4
            val = parse_range(val)
        elif key == 'content-range':
            flags |= 8
            val = parse_content_range(val)
        else:
            val = _parse_val(val)
        if headers.setdefault(key, val) is not val:
            raise ValueError(
                'duplicate header: {!r}'.format(line)
            )
    if (flags & 3) == 3:
        raise ValueError(
            'cannot have both content-length and transfer-encoding headers'
        )
    if (flags & 4):
        if (flags & 3):
            raise ValueError(
                'cannot include range header and content-length/transfer-encoding'
            )
        if isresponse:
            raise ValueError(
                "response cannot include a 'range' header"
            )
    if (flags & 8) and not isresponse:
        raise ValueError(
            "request cannot include a 'content-range' header"
        )
    return headers


def parse_headers(src, isresponse=False):
    if src == b'':
        return {}
    return _parse_header_lines(src.split(b'\r\n'), isresponse)



################################################################################
# Request parsing:

def _parse_method(src):
    assert isinstance(src, bytes)
    method = _METHODS.get(src)
    if method is None:
        raise ValueError('bad HTTP method: {!r}'.format(src))
    return method


def parse_method(src):
    if isinstance(src, str):
        src = src.encode()
    return _parse_method(src)


def _parse_path_component(src):
    if PATH.issuperset(src):
        return src.decode('ascii')
    raise ValueError('bad bytes in path component: {!r}'.format(src))


def _parse_path(src):
    if not src:
        raise ValueError('path is empty')
    if src[0:1] != b'/':
        raise ValueError("path[0:1] != b'/': {!r}".format(src))
    if b'//' in src:
        raise ValueError("b'//' in path: {!r}".format(src))
    if src == b'/':
        return []
    return [_parse_path_component(c) for c in src[1:].split(b'/')]


def _parse_query(src):
    if QUERY.issuperset(src):
        return src.decode('ascii')
    raise ValueError('bad bytes in query: {!r}'.format(src))


def _parse_uri(src):
    if not src:
        raise ValueError('uri is empty')
    if not URI.issuperset(src):
        raise ValueError('bad bytes in uri: {!r}'.format(src))
    parts = src.split(b'?', 1)
    path = _parse_path(parts[0])
    if len(parts) == 1:
        return (path, None)
    return (path, _parse_query(parts[1]))


def parse_uri(src):
    (path, query) = _parse_uri(src)
    return (src.decode('ascii'), [], path, query)


def parse_request_line(line):
    if len(line) < 14:
        raise ValueError('request line too short: {!r}'.format(line))
    if line[-9:] != b' HTTP/1.1':
        raise ValueError('bad protocol in request line: {!r}'.format(line[-9:]))
    src = line[:-9]
    items = src.split(b' /', 1)
    if len(items) < 2:
        raise ValueError('bad request line: {!r}'.format(line))
    method = _parse_method(items[0])
    (uri, mount, path, query) = parse_uri(b'/' + items[1])
    return (method, uri, mount, path, query)


def parse_request(preamble, rfile):
    if preamble == b'':
        raise EmptyPreambleError('request preamble is empty')
    (first_line, *header_lines) = preamble.split(b'\r\n')
    (method, uri, mount, path, query) = parse_request_line(first_line)
    headers = _parse_header_lines(header_lines)
    if 'content-length' in headers:
        body = Body(rfile, headers['content-length'])
    elif 'transfer-encoding' in headers:
        body = ChunkedBody(rfile)
    else:
        body = None
    return Request(method, uri, headers, body, mount, path, query)


_REQUEST_REPR = 'Request(method={!r}, uri={!r}, headers={!r}, body={!r}, \
mount={!r}, path={!r}, query={!r})'


class Request:
    __slots__ = (
        '_method', '_uri', '_headers', '_body', '_mount', '_path', '_query'
    )

    def __init__(self, method, uri, headers, body, mount, path, query):
        self._method = method
        self._uri = uri
        self._headers = headers
        self._body = body
        self._mount = _check_list('mount', mount)
        self._path = _check_list('path', path)
        self._query = query

    def __repr__(self):
        return _REQUEST_REPR.format(
            self._method,
            self._uri,
            self._headers,
            self._body,
            self._mount,
            self._path,
            self._query,
        )

    @property
    def method(self):
        return self._method

    @property
    def uri(self):
        return self._uri

    @property
    def headers(self):
        return self._headers

    @property
    def body(self):
        return self._body

    @property
    def mount(self):
        return self._mount

    @property
    def path(self):
        return self._path

    @property
    def query(self):
        return self._query

    def shift_path(self):
        if not self._path:
            raise IndexError('Request.path is empty')
        next = self._path.pop(0)
        self._mount.append(next)
        return next

    def build_proxy_uri(self):
        uri = '/' + '/'.join(self._path)
        if self._query is None:
            return uri
        return '?'.join([uri, self._query])



################################################################################
# Response parsing:

def _parse_status(src):
    if DECIMAL.issuperset(src):
        status = int(src)
        if 100 <= status <= 599:
            return status
    raise ValueError('bad status: {!r}'.format(src))


def _parse_reason(src):
    if REASON.issuperset(src):
        if src == b'OK':
            return _OK
        return src.decode('ascii')
    raise ValueError('bad reason: {!r}'.format(src))


def parse_response_line(src):
    assert isinstance(src, bytes)
    if len(src) < 15:
        raise ValueError('response line too short: {!r}'.format(src))
    if src[0:9] != b'HTTP/1.1 ' or src[12:13] != b' ':
        raise ValueError('bad response line: {!r}'.format(src))
    status = _parse_status(src[9:12])
    reason = _parse_reason(src[13:])
    return (status, reason)


def parse_response(method, preamble, rfile):
    _check_method(method)
    if preamble == b'':
        raise EmptyPreambleError('response preamble is empty')
    (first_line, *header_lines) = preamble.split(b'\r\n')
    (status, reason) = parse_response_line(first_line)
    headers = _parse_header_lines(header_lines, isresponse=True)
    if method == 'HEAD':
        body = None
    elif 'content-length' in headers:
        body = Body(rfile, headers['content-length'])
    elif 'transfer-encoding' in headers:
        body = ChunkedBody(rfile)
    else:
        body = None
    return Response(status, reason, headers, body)


################################################################################
# Rendering and formatting:

def _check_method(method):
    _check_str('method', method)
    if method not in {'GET', 'PUT', 'POST', 'HEAD', 'DELETE'}:
        raise ValueError('bad method: {!r}'.format(method))

def _validate_chunk(chunk):
    _check_tuple('chunk', chunk, 2)
    (ext, data) = chunk
    if ext is not None:
        _check_tuple('chunk[0]', ext, 2)
    _check_bytes('chunk[1]', data, MAX_IO_SIZE)
    return chunk


def _format_chunk(size, ext):
    if ext is None:
        return '{:x}\r\n'.format(size).encode()
    (key, value) = ext
    return '{:x};{}={}\r\n'.format(size, key, value).encode()


def format_chunk(chunk):
    (ext, data) = _validate_chunk(chunk)
    if ext is None:
        return '{:x}\r\n'.format(len(data)).encode()
    (key, value) = ext
    return '{:x};{}={}\r\n'.format(len(data), key, value).encode()


def _write_chunk(wobj, chunk):
    line = format_chunk(chunk)
    total =  _write_to(wobj, line)
    total += _write_to(wobj, chunk[1])
    total += _write_to(wobj, b'\r\n')
    return total


def write_chunk(wfile, chunk):
    return _write_chunk(_get_wobj(wfile), chunk)


class _Output:
    __slots__ = ('dst', 'stop')

    def __init__(self, dst):
        assert isinstance(dst, memoryview)
        self.dst = dst
        self.stop = 0

    def copy_into(self, src):
        assert type(src) is bytes
        assert self.stop <= len(self.dst)
        if self.stop + len(src) > len(self.dst):
            raise ValueError(
                'output size exceeds {}'.format(len(self.dst))
            )
        start = self.stop
        stop = start + len(src)
        self.dst[start:stop] = src
        self.stop += len(src)


_OUTGOING_KEY = frozenset('-0123456789abcdefghijklmnopqrstuvwxyz')

def _check_key(key):
    _check_type('key', key, str)
    if len(key) > SCRATCH_LEN:
        raise ValueError('key is too long: {!r}'.format(key))
    if key == '' or not _OUTGOING_KEY.issuperset(key):
        raise ValueError(
            'bad key: {!r}'.format(key)
        )
    return key


def _check_val(val):
    if type(val) is not str:
        val = str(val)
    return _check_str('val', val)


def _render_headers(o, headers):
    _check_type('headers', headers, dict)
    if len(headers) > MAX_HEADER_COUNT:
        raise ValueError(
            'need len(headers) <= {}; got {}'.format(
                MAX_HEADER_COUNT, len(headers)
            )
        )
    for key in headers:
        _check_key(key)
    src = ''.join(
        '\r\n{}: {}'.format(k, _check_val(v))
        for (k, v) in sorted(headers.items()) 
    ).encode('ascii')
    o.copy_into(src)


def render_headers(dst, headers):
    o = _Output(dst)
    _render_headers(o, headers)
    return o.stop


def _render_request(o, method, uri, headers):
    _check_str('method', method)
    _check_str('uri', uri)
    line = '{} {} HTTP/1.1'.format(method, uri).encode('ascii')
    o.copy_into(line)
    _render_headers(o, headers)
    o.copy_into(b'\r\n\r\n')


def render_request(dst, method, uri, headers):
    o = _Output(dst)
    _render_request(o, method, uri, headers)
    return o.stop


def _render_response(o, status, reason, headers):
    _check_int('status', status, 100, 599)
    _check_str('reason', reason)
    line = 'HTTP/1.1 {} {}'.format(status, reason).encode('ascii')
    o.copy_into(line)
    _render_headers(o, headers)
    o.copy_into(b'\r\n\r\n')


def render_response(dst, status, reason, headers):
    o = _Output(dst)
    _render_response(o, status, reason, headers)
    return o.stop


################################################################################
# Reader:

class Reader:
    __slots__ = (
        '_sock_recv_into',
        '_rawtell',
        '_rawbuf',
        '_start',
        '_buf',
        '_closed',
    )

    def __init__(self, sock):
        self._sock_recv_into = _getcallable('sock', sock, 'recv_into')
        self._rawtell = 0
        self._rawbuf = memoryview(bytearray(BUF_LEN))
        self._start = 0
        self._buf = b''
        self._closed = False

    def rawtell(self):
        return self._rawtell

    def tell(self):
        return self._rawtell - len(self._buf)

    def _recv_into(self, buf):
        added = self._sock_recv_into(buf)
        _validate_size('received', added, len(buf))
        self._rawtell += added
        return added

    def _update(self, start, size):
        """
        Valid transitions::

            ===========================
            -->|<--            |  Empty
               |<---- buf <--  |  Shift
               |      buf ---->|  Fill
               |  --> buf      |  Drain
            ===========================
        """
        # Check previous state:
        assert 0 <= self._start <= self._start + len(self._buf) <= len(self._rawbuf)

        # Check new state:
        assert 0 <= start <= start + size <= len(self._rawbuf)

        # _update() should only be called when there is a change:
        assert start != self._start or size != len(self._buf)

        # Check that previous to new is one of the four valid transitions:
        if size == 0:
            # empty
            assert start == 0
            assert len(self._buf) > 0
        elif size == len(self._buf):
            # shift
            assert size > 0
            assert start == 0
            assert self._start > 0
        elif size > len(self._buf):
            # fill
            assert size > 0
            assert start == self._start == 0
        elif size < len(self._buf):
            # drain
            assert size > 0
            assert start + size == self._start + len(self._buf)
        else:
            raise ValueError(
                'invalid buffer update: ({},{}) --> ({}, {})'.format(
                    self._start, len(self._buf), start, size
                )
            )

        # Update start, buf:
        self._start = start
        self._buf = self._rawbuf[start:start+size].tobytes()

    def expose(self):
        return self._rawbuf.tobytes()

    def peek(self, size):
        assert isinstance(size, int)
        if size < 0:
            return self._buf
        return self._buf[0:size]

    def _drain(self, size):
        avail = len(self._buf)
        src = self.peek(size)
        if len(src) == 0:
            return src
        if len(src) == avail:
            self._update(0, 0)
        else:
            self._update(self._start + len(src), avail - len(src))
        return src

    def _found(self, index, end):
        src = self._drain(index + len(end))
        return src[0:-len(end)]

    def _not_found(self, cur, end):
        if len(cur) == 0:
            return cur
        raise ValueError(
            '{!r} not found in {!r}...'.format(end, cur[:32])
        )

    def read_until(self, size, end):
        end = memoryview(end).tobytes()
        assert type(size) is int

        if not end:
            raise ValueError('end cannot be empty')
        if not (len(end) <= size <= len(self._rawbuf)):
            raise ValueError(
                'need {} <= size <= {}; got {}'.format(
                    len(end), len(self._rawbuf), size
                )
            )

        # First, search current buffer:
        cur = self.peek(size)
        index = cur.find(end)
        if index >= 0:
            return self._found(index, end)
        if len(cur) == size:
            return self._not_found(cur, end)

        # Shift buffer if needed:
        if self._start > 0:
            assert len(cur) > 0
            self._rawbuf[0:len(cur)] = cur
            self._update(0, len(cur))

        # Now search till found:
        start = len(cur)
        while start < size:
            dst = self._rawbuf[start:]
            added = self._recv_into(dst)
            if added == 0:
                break
            start += added
            self._update(0, start)
            cur = self.peek(size)
            index = cur.find(end)
            if index >= 0:
                return self._found(index, end)

        # Didn't find it:
        return self._not_found(cur, end)

    def read_request(self):
        preamble = self.read_until(len(self._rawbuf), b'\r\n\r\n')
        return parse_request(preamble, self)

    def read_response(self, method):
        _check_method(method)
        preamble = self.read_until(len(self._rawbuf), b'\r\n\r\n')
        return parse_response(method, preamble, self)

    def readchunkline(self):
        line = self.read_until(4096, b'\r\n')
        return parse_chunk(line)

    def readinto(self, dst):
        dst = memoryview(dst)
        dst_len = len(dst)
        if not (1 <= dst_len <= MAX_IO_SIZE):
            raise ValueError(
                'need 1 <= len(buf) <= {}; got {}'.format(MAX_IO_SIZE, dst_len)
            )
        src = self._drain(dst_len)
        src_len = len(src)
        dst[0:src_len] = src
        added = _readinto(self._sock_recv_into, dst[src_len:])
        assert added is not None
        self._rawtell += added
        assert dst_len == src_len + added
        return dst_len


################################################################################
# Writer:

def set_default_header(headers, key, val):
    assert isinstance(headers, dict)
    assert isinstance(key, str)
    cur = headers.setdefault(key, val)
    if val != cur:
        raise ValueError(
            '{!r} mismatch: {!r} != {!r}'.format(key, val, cur)
        )


def set_output_headers(headers, body):
    if body is None:
        return
    if type(body) is bytes:
        set_default_header(headers, 'content-length', len(body))
    elif type(body) in (Body, BodyIter):
        set_default_header(headers, 'content-length', body.content_length)
    elif type(body) in (ChunkedBody, ChunkedBodyIter):
        set_default_header(headers, 'transfer-encoding', 'chunked')
    else:
        raise TypeError(
            'bad body type: {!r}: {!r}'.format(type(body), body)
        )
    


class Writer:
    __slots__ = (
        '_sock_send',
        '_tell',
        '_buf',
        '_stop',
    )

    def __init__(self, sock):
        self._sock_send = _getcallable('sock', sock, 'send')
        self._tell = 0
        self._buf = memoryview(bytearray(BUF_LEN))
        self._stop = 0

    def tell(self):
        return self._tell

    def _raw_write(self, src):
        size = _write(self._sock_send, src)
        assert size == len(src)
        self._tell += size
        return size

    def _flush(self):
        assert 0 <= self._stop <= len(self._buf)
        src = self._buf[0:self._stop]
        self._raw_write(src)
        self._stop = 0

    def _write(self, src):
        if self._stop > 0 and self._stop + len(src) <= len(self._buf):
            start = self._stop
            self._stop += len(src)
            self._buf[start:self._stop] = src
            self._flush()
            return len(src)
        self._flush()
        return self._raw_write(src)

    def _write_output(self, func, arg1, arg2, headers, body):
        orig_tell = self._tell
        assert self._stop == 0

        total = func(self._buf, arg1, arg2, headers)
        assert total > 0
        self._stop = total
        if type(body) is bytes:
            if len(body) > MAX_IO_SIZE:
                raise ValueError(
                    'need len(body) <= {}; got {}'.format(
                        MAX_IO_SIZE, len(body)
                    )
                )
            total += self._write(body)
        elif type(body) in bodies:
            total += body.write_to(self)
        elif body is not None:
            raise TypeError(
                'bad body type: {!r}: {!r}'.format(type(body), body)
            )
        self._flush()
        assert self._stop == 0
        assert self._tell == orig_tell + total
        return total

    def write_request(self, method, uri, headers, body):
        _check_method(method)
        set_output_headers(headers, body)
        return self._write_output(render_request,
            method, uri, headers, body
        )

    def write_response(self, status, reason, headers, body):
        set_output_headers(headers, body)
        return self._write_output(render_response,
            status, reason, headers, body
        )

def _check_body_state(name, state, max_state):
    assert max_state < BODY_CONSUMED
    if state <= max_state:
        return
    if state is BODY_STARTED:
        raise ValueError(
            '{}.state == BODY_STARTED, cannot start another operation'.format(
                name
            )
        )
    if state is BODY_CONSUMED:
        raise ValueError(
            '{}.state == BODY_CONSUMED, already consumed'.format(name)
        )
    if state is BODY_ERROR:
        raise ValueError(
            '{}.state == BODY_ERROR, cannot be used'.format(name)
        )
    raise Exception('bad state: {!r}'.format(state))


def _rfile_repr(rfile):
    if type(rfile) is Reader:
        return '<reader>'
    return '<rfile>'


class Body:
    __slots__ = (
        '_rfile',
        '_robj',
        '_content_length',
        '_remaining',
        '_state',
        '_chunked',
    )

    def __init__(self, rfile, content_length):
        _validate_length('content_length', content_length)
        self._rfile = rfile
        self._remaining = self._content_length = content_length
        self._state = BODY_READY
        if type(rfile) is Reader:
            self._robj = rfile
        else:
            self._robj = _getcallable('rfile', rfile, 'readinto')
        self._chunked = False

    @property
    def rfile(self):
        return self._rfile

    @property
    def content_length(self):
        return self._content_length

    @property
    def state(self):
        return self._state

    @property
    def chunked(self):
        return self._chunked

    def __repr__(self):
        return 'Body({}, {!r})'.format(
            _rfile_repr(self._rfile), self._content_length
        )

    def __iter__(self):
        _check_body_state('Body', self._state, BODY_READY)
        self._state = BODY_STARTED
        try:
            remaining = self._remaining
            iosize = min(remaining, IO_SIZE)
            dst = memoryview(bytearray(iosize))
            robj = self._robj
            while remaining > 0:
                size = min(remaining, iosize)
                remaining -= size
                assert remaining >= 0
                sub = dst[:size]
                _readinto_from(robj, sub)
                yield sub.tobytes()
        except:
            self._state = BODY_ERROR
            raise
        assert remaining == 0
        self._remaining = remaining
        self._state = BODY_CONSUMED

    def read(self, size=None):
        rsize = _validate_read_size('size', size, self._remaining)
        _check_body_state('Body', self._state, BODY_STARTED)
        self._state = BODY_STARTED
        if self._remaining == 0:
            self._state = BODY_CONSUMED
            return b''
        try:
            rsize = min(self._remaining, rsize)
            dst = memoryview(bytearray(rsize))
            _readinto_from(self._robj, dst)
            self._remaining -= rsize
            assert self._remaining >= 0
            if size is None:
                self._state = BODY_CONSUMED
            return dst.tobytes()
        except:
            self._state = BODY_ERROR
            raise

    def write_to(self, wfile):
        wobj = _get_wobj(wfile)
        total = sum(_write_to(wobj, data) for data in self)
        assert total == self._content_length
        return total

def _not_found(self, cur, end, readline):
    if readline:
        return self._drain(len(cur))
    if len(cur) == 0:
        return cur
    raise ValueError(
        '{!r} not found in {!r}...'.format(end, cur[:32])
    )


def _readchunk(readline, read):
    line = readline(4096)
    if type(line) is not bytes:
        raise TypeError(
            'need a {!r}; readline() returned a {!r}'.format(bytes, type(line))
        )
    if len(line) > 4096:
        raise ValueError(
            'readline() returned too many bytes: {} > {}'.format(len(line), 4096)
        )
    if line[-2:] != b'\r\n':
        raise ValueError(
            '{!r} not found in {!r}...'.format(b'\r\n', line[:32])
        )
    (size, ext) = parse_chunk(line[:-2])
    data = read(size + 2)
    if type(data) is not bytes:
        raise TypeError(
            'need a {!r}; read() returned a {!r}'.format(bytes, type(data))
        )
    if len(data) != size + 2:
        raise ValueError(
            'read() returned {} bytes, need {}'.format(len(data), size + 2)
        )
    end = data[-2:]
    if end != b'\r\n':
        raise ValueError('bad chunk data termination: {!r}'.format(end))
    return (ext, data)
    

def _readchunkline(readline):
    line = readline(4096)
    if type(line) is not bytes:
        raise TypeError(
            'need a {!r}; readline() returned a {!r}'.format(bytes, type(line))
        )
    if len(line) > 4096:
        raise ValueError(
            'readline() returned too many bytes: {} > {}'.format(len(line), 4096)
        )
    if line[-2:] != b'\r\n':
        raise ValueError(
            '{!r} not found in {!r}...'.format(b'\r\n', line[:32])
        )
    return parse_chunk(line[:-2])


def _readchunk_from(robj, readline, nopack=False):
    if type(robj) is Reader:
        (size, ext) = robj.readchunkline()
    else:
        (size, ext) = _readchunkline(readline)
    dst = memoryview(bytearray(size + 2))
    _readinto_from(robj, dst)
    if dst[-2:] != b'\r\n':
        raise ValueError(
            'bad chunk data termination: {!r}'.format(dst[-2:].tobytes())
        )
    if nopack:
        return (size, ext, dst)
    return (ext, dst[:-2].tobytes())


def readchunk(rfile):
    robj = _get_robj(rfile)
    readline = _get_readline(rfile)
    return _readchunk_from(robj, readline)


class ChunkedBody:
    __slots__ = ('_rfile', '_robj', '_readline', '_state', '_chunked')

    def __init__(self, rfile):
        self._rfile = rfile
        self._robj = _get_robj(rfile)
        self._readline = _get_readline(rfile)
        self._state = BODY_READY
        self._chunked = True

    def __repr__(self):
        return 'ChunkedBody({})'.format(_rfile_repr(self._rfile))

    @property
    def rfile(self):
        return self._rfile

    @property
    def state(self):
        return self._state

    @property
    def chunked(self):
        return self._chunked

    def readchunk(self):
        _check_body_state('ChunkedBody', self._state, BODY_STARTED)
        self._state = BODY_STARTED
        try:
            chunk = _readchunk_from(self._robj, self._readline)
            if len(chunk[1]) == 0:
                self._state = BODY_CONSUMED
        except:
            self._state = BODY_ERROR
            raise
        return chunk

    def read(self):
        _check_body_state('ChunkedBody', self._state, BODY_STARTED)
        self._state = BODY_STARTED
        try:
            total = 0
            accum = []
            while total <= MAX_IO_SIZE:
                (ext, data) = self.readchunk()
                total += len(data)
                if len(data) == 0:
                    break
                accum.append(data)
            if total > MAX_IO_SIZE:
                raise ValueError(
                    'chunks exceed MAX_IO_SIZE: {} > {}'.format(
                        total, MAX_IO_SIZE
                    )
                )
            ret =  b''.join(accum)
        except:
            self._state = BODY_ERROR
            raise
        self._state = BODY_CONSUMED
        return ret

    def __iter__(self):
        _check_body_state('ChunkedBody', self._state, BODY_READY)
        self._state = BODY_STARTED
        while self._state < BODY_CONSUMED:
            yield self.readchunk()

    def write_to(self, wfile):
        _check_body_state('ChunkedBody', self._state, BODY_READY)
        self._state = BODY_STARTED
        robj = self._robj
        readline = self._readline
        wobj = _get_wobj(wfile)
        readchunk_from = _readchunk_from
        format_chunk = _format_chunk
        write_to = _write_to
        total = 0
        try:
            while True:
                (size, ext, data) = readchunk_from(robj, readline, nopack=True)
                assert len(data) == size + 2 and data[-2:] == b'\r\n'
                line = format_chunk(size, ext)
                total += write_to(wobj, line)
                total += write_to(wobj, data)
                if size == 0:
                    break
        except:
            self._state = BODY_ERROR
            raise
        self._state = BODY_CONSUMED
        return total


class BodyIter:
    __slots__ = ('_source', '_state', '_content_length')

    def __init__(self, source, content_length):
        self._source = source
        self._state = BODY_READY
        self._content_length = _validate_length("content_length", content_length)

    def __repr__(self):
        return 'BodyIter(<source>, {})'.format(self._content_length)

    @property
    def source(self):
        return self._source

    @property
    def content_length(self):
        return self._content_length

    @property
    def state(self):
        return self._state

    def write_to(self, wfile):
        _check_body_state('BodyIter', self._state, BODY_READY)
        self._state = BODY_STARTED
        wobj = _get_wobj(wfile)
        length = self._content_length
        total = 0
        try:
            for part in self._source:
                if type(part) is not bytes:
                    _check_bytes('BodyIter source item', part)
                total += _write_to(wobj, part)
                if total > length:
                    raise ValueError(
                        'exceeds content_length: {} > {}'.format(total, length)
                    )
            if total != length:
                raise ValueError(
                    'deceeds content_length: {} < {}'.format(total, length)
                )
        except:
            self._state = BODY_ERROR
            raise
        self._state = BODY_CONSUMED
        return total


class ChunkedBodyIter:
    __slots__ = ('_source', '_state')

    def __init__(self, source):
        self._source = source
        self._state = BODY_READY

    def __repr__(self):
        return 'ChunkedBodyIter(<source>)'

    @property
    def source(self):
        return self._source

    @property
    def state(self):
        return self._state

    def write_to(self, wfile):
        _check_body_state('ChunkedBodyIter', self._state, BODY_READY)
        self._state = BODY_STARTED
        wobj = _get_wobj(wfile)
        empty = False
        total = 0
        try:
            for chunk in self._source:
                if empty:
                    raise ValueError('additional chunk after empty chunk data')
                total += _write_chunk(wobj, chunk)
                if not chunk[1]:  # Is chunk data empty?
                    empty = True
            if not empty:
                raise ValueError('final chunk data was not empty')
        except:
            self._state = BODY_ERROR
            raise
        self._state = BODY_CONSUMED
        return total

# Used to expose the RGI IO wrappers:
bodies = Bodies(Body, ChunkedBody, BodyIter, ChunkedBodyIter)


def _check_dict(name, obj):
    if type(obj) is not dict:
        raise TypeError(
            TYPE_ERROR.format(name, dict, type(obj), obj)
        )


def _body_is_consumed(body):
    if body is None:
        return True
    assert type(body) in (Body, ChunkedBody)
    return body._state == BODY_CONSUMED


def _unpack_response(obj):
    if type(obj) is Response:
        return obj
    return _check_tuple('response', obj, 4)   

def _check_status(status):
    _validate_int('status', status)
    if not 100 <= status <= 599:
        raise ValueError(
            'need 100 <= status <= 599; got {}'.format(status)
        )

def _check_int(name, obj, _min, _max):
    _validate_int(name, obj)
    if not _min <= obj <= _max:
        raise ValueError(
            'need {} <= {} <= {}; got {}'.format(_min, name, _max, obj)
        )
    return obj


class Session:
    __slots__ = (
        '_address',
        '_credentials',
        '_max_requests',
        '_requests',
        '_store',
        '_closed',
        '_message',
    )

    def __init__(self, address, credentials=None, max_requests=None):
        if credentials is not None:
            _check_tuple('credentials', credentials, 3)
        if max_requests is None:
            max_requests=500
        self._address = address
        self._credentials = credentials
        self._max_requests = _check_int('max_requests', max_requests, 0, 75000)
        self._requests = 0
        self._store = {}
        self._closed = False
        self._message = None

    def __repr__(self):
        if self._credentials is None:
            return 'Session({!r})'.format(self._address)
        return 'Session({!r}, {!r})'.format(self._address, self._credentials)

    def __str__(self):
        if self._credentials is None:
            return repr(self._address)
        return '{!r} {!r}'.format(self._address, self._credentials)

    @property
    def address(self):
        return self._address

    @property
    def credentials(self):
        return self._credentials

    @property
    def max_requests(self):
        return self._max_requests

    @property
    def requests(self):
        return self._requests

    @property
    def store(self):
        return self._store

    @property
    def closed(self):
        return self._closed

    @property
    def message(self):
        return self._message

    def close(self, message):
        self._closed = True
        if self._message is None:
            self._message = message

    def _response_complete(self, status, reason):
        if status >= 400 and status not in (404, 409, 412):
            self.close('{} {}'.format(status, reason))
        self._requests += 1
        if self._requests >= self._max_requests:
            self.close('max_requests')


def _handle_requests(app, session, sock):
    _check_type2('session', session, Session)
    assert session.requests == session._requests == 0
    reader = Reader(sock)
    writer = Writer(sock)
    while not session._closed:
        request = reader.read_request()
        response = app(session, request, bodies)
        (status, reason, headers, body) = _unpack_response(response)
        _check_status(status)

        # Make sure application fully consumed request body:
        if not _body_is_consumed(request.body):
            raise ValueError(
                'request body not consumed: {!r}'.format(request.body)
            )

        # FIXME: when 200 <= status <= 299, we should consider requiring that
        # the response body for a GET request not be None

        # Make sure HEAD requests are properly handled:
        if request.method == 'HEAD' and body is not None:
            raise TypeError(
                'request method is HEAD, but response body is not None: {!r}'.format(
                    type(body)
                )
            )
            # FIXME: when 200 <= status <= 299, we should consider requiring
            # a 'content-length' or 'transfer-encoding' header in the response
            # to a HEAD request

        # Write response:
        writer.write_response(status, reason, headers, body)

        # Update request counter, possibly close based on status:
        session._response_complete(status, reason)


class Connection:
    __slots__ = (
        'sock',
        'base_headers',
        '_reader',
        '_writer',
        '_response_body',
        '_closed',
        '_bodies',
    )

    def __init__(self, sock, base_headers):
        self._closed = False
        self.sock = sock
        if base_headers is not None:
            _check_dict('base_headers', base_headers)
        self.base_headers = base_headers
        self._reader = Reader(sock)
        self._writer = Writer(sock)
        self._response_body = None
        self._closed = False
        self._bodies = bodies

    @property
    def closed(self):
        return self._closed

    @property
    def bodies(self):
        return self._bodies

    def __del__(self):
        self._shutdown()

    def _shutdown(self, how=socket.SHUT_RDWR):
        if self._closed is not True:
            self._closed = True
            try:
                self.sock.shutdown(how)
            except:
                pass
            try:
                self.sock.close()
            except:
                pass

    def close(self):
        self._shutdown()

    def request(self, method, uri, headers, body):
        if self._closed is not False:
            raise ValueError('Connection is closed')
        try:
            _check_method(method)
            if body is not None and method not in {'PUT', 'POST'}:
                raise ValueError(
                    'when method is {!r}, body must be None; got a {!r}'.format(
                        method, type(body)
                    )
                )
            if not _body_is_consumed(self._response_body):
                raise ValueError(
                    'response body not consumed: {!r}'.format(self._response_body)
                )
            if self.base_headers:
                headers.update(self.base_headers)
            self._writer.write_request(method, uri, headers, body)
            response = self._reader.read_response(method)
            self._response_body = response.body
            return response
        except Exception:
            self._shutdown()
            raise

    def put(self, uri, headers, body):
        return self.request('PUT', uri, headers, body)

    def post(self, uri, headers, body):
        return self.request('POST', uri, headers, body)

    def get(self, uri, headers):
        return self.request('GET', uri, headers, None)

    def head(self, uri, headers):
        return self.request('HEAD', uri, headers, None)

    def delete(self, uri, headers):
        return self.request('DELETE', uri, headers, None)

    def get_range(self, uri, headers, start, stop):
        set_default_header(headers, 'range', Range(start, stop))
        return self.request('GET', uri, headers, None)




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
A few simple utility functions useful for most RGI server applications.

The module is heavily inspired by the `wsgiref.util` module in the Python3
standard library:

    https://docs.python.org/3/library/wsgiref.html

"""


def shift_path(request):
    """
    Shift path to mount in an RGI *request* argument.

    This function only uses the `mount` and `path` attributes from the
    *request* namedtuple:

    >>> from collections import namedtuple
    >>> Request = namedtuple('Request', 'mount path')
    >>> request = Request(['foo'], ['bar', 'baz'])
    >>> shift_path(request)
    'bar'

    And you can see *request* was updated in place:

    >>> request.mount
    ['foo', 'bar']
    >>> request.path
    ['baz']

    """
    next = request.path.pop(0)
    request.mount.append(next)
    return next


def relative_uri(request):
    """
    Reconstruct a relative URI from an RGI *request* argument.

    This function uses the `path` and `query` attributes from the *request*
    namedtuple:

    >>> from collections import namedtuple
    >>> Request = namedtuple('Request', 'path query')

    For example, when there is no query:

    >>> request = Request(['bar', 'baz'], None)
    >>> relative_uri(request)
    '/bar/baz'

    Or when there is merely an empty query:

    >>> request = Request(['bar', 'baz'], '')
    >>> relative_uri(request)
    '/bar/baz?'

    And when there is a query:

    >>> request = Request(['bar', 'baz'], 'stuff=junk')
    >>> relative_uri(request)
    '/bar/baz?stuff=junk'

    Note that ``request.mount`` is ignored by this function.
    """
    uri = '/' + '/'.join(request.path)
    if request.query is None:
        return uri
    return '?'.join([uri, request.query])


def absolute_uri(request):
    """
    Reconstruct an absolute URI from an RGI *request* argument.

    This function uses the `mount`, `path`, and `query` attributes from the
    *request* namedtuple:

    >>> from collections import namedtuple
    >>> Request = namedtuple('Request', 'mount path query')

    For example, when there is no query:

    >>> request = Request(['foo'], ['bar', 'baz'], None)
    >>> absolute_uri(request)
    '/foo/bar/baz'

    Or when there is merely an empty query:

    >>> request = Request(['foo'], ['bar', 'baz'], '')
    >>> absolute_uri(request)
    '/foo/bar/baz?'

    And when there is a query:

    >>> request = Request(['foo'], ['bar', 'baz'], 'stuff=junk')
    >>> absolute_uri(request)
    '/foo/bar/baz?stuff=junk'

    """
    uri = '/' + '/'.join(request.mount + request.path)
    if request.query is None:
        return uri
    return '?'.join([uri, request.query])




#!/usr/bin/python3

"""
Benchmark file transfer rate over SSL.

This benchmark simulates file-transfer as used by Dmedia, a scenario where there
is little HTTP overhead (because the files are large), and also little concern
for the cost of creating the SSL connection (both because the files are large,
and also because Dmedia will typically request many files through the same
connection).

Our goal is to be able to saturate a 10 gigabit local Ethernet connection with a
single connection, and a single CPU core.  On Haswell processors, we're already
very close to this as the GCM modes (authenticated encryption) in TLS 1.2 are
very fast.

Example results on an i7-4900MQ (2.8 GHz, 64-bit Ubuntu 14.04, Python 3.4)::

    cipher: ('ECDHE-RSA-AES256-GCM-SHA384', 'TLSv1/SSLv3', 256)
    runs: [2.8518437089987856, 2.8778152399991086, 2.858511821999855]
    fastest run: 2.8518437089987856
    transfer rate in fastest run: 941 MB/s

That's 7528 M bit/s, so we're about 75% the way there.  Of course, with multiple
parallel connections on multiple cores, we can fully saturate 10 gigabit today.

Even on 64-bit systems, there is a small performance improvement with AES-128,
but perhaps not enough to justify the switch (something to consider though)::

    cipher: ('ECDHE-RSA-AES128-GCM-SHA256', 'TLSv1/SSLv3', 128)
    runs: [2.647685547000947, 2.6391498539996974, 2.6463401510009135]
    fastest run: 2.6391498539996974
    transfer rate in fastest run: 1.02 GB/s
"""

import time
import logging
import os
import math

from degu import IPv6_LOOPBACK
from degu.misc import TempPKI, TempSSLServer
from degu.client import SSLClient


logging.basicConfig(
    level=logging.DEBUG,
    format='\t'.join([
        '%(levelname)s',
        '%(threadName)s',
        '%(message)s',
    ]),
)


BYTES10 = (
    'bytes',
    'kB',
    'MB',
    'GB',
    'TB',
    'PB',
    'EB',
    'ZB',
    'YB',
)


def bytes10(size):
    if size is None:
        return None
    if size < 0:
        raise ValueError('size must be >= 0; got {!r}'.format(size))
    if size == 0:
        return '0 bytes'
    if size == 1:
        return '1 byte'
    i = min(int(math.floor(math.log(size, 1000))), len(BYTES10) - 1)
    s = (size / (1000 ** i) if i > 0 else size)
    return (
        '{:.3g} {}'.format(s, BYTES10[i])
    )


MiB = 1048576
chunk_size = MiB
unique_chunks = tuple(os.urandom(chunk_size) for i in range(8))
chunks = unique_chunks * 64  # 512 MiB 
content_length = sum(len(c) for c in chunks)


def file_app(session, request, bodies):
    headers = {'content-type': 'video/quicktime'}
    body = bodies.BodyIter(chunks, content_length)
    return (200, 'OK', headers, body)


pki = TempPKI()
server = TempSSLServer(pki.server_sslconfig, IPv6_LOOPBACK, file_app)
client = SSLClient(pki.client_sslconfig, server.address)


count = 5
deltas = []
for i in range(5):
    conn = client.connect()
    start = time.monotonic()
    for i in range(count):
        r = conn.request('GET', '/', {}, None)
        cipher = conn.sock.cipher()
        for block in r.body:
            pass
    deltas.append(time.monotonic() - start)
    conn.close()
server.terminate()
fastest = min(deltas)
rate = (content_length * count) / fastest
print('')
print('cipher: {}'.format(cipher))
print('runs: {}'.format(deltas))
print('fastest run: {}'.format(fastest))
print('transfer rate in fastest run: {}/s'.format(bytes10(int(rate))))




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for the `degu` package.
"""

from unittest import TestCase
import os
import multiprocessing

import degu
from degu.client import Client, SSLClient
from degu.misc import TempPKI


class TestConstants(TestCase):
    def test_version(self):
        self.assertIsInstance(degu.__version__, str)
        parts = degu.__version__.split('.')
        self.assertEqual(len(parts), 3)
        for part in parts:
            p = int(part)
            self.assertTrue(p >= 0)
            self.assertEqual(str(p), part)

    def test_ADDRESS_CONSTANTS(self):
        self.assertIsInstance(degu.ADDRESS_CONSTANTS, tuple)
        self.assertEqual(degu.ADDRESS_CONSTANTS, (
            degu.IPv6_LOOPBACK,
            degu.IPv6_ANY,
            degu.IPv4_LOOPBACK,
            degu.IPv4_ANY,
        ))
        for address in degu.ADDRESS_CONSTANTS:
            self.assertIsInstance(address, tuple)
            self.assertIn(len(address), {2, 4})
            self.assertIsInstance(address[0], str)
            self.assertIn(address[0], {'::1', '::', '127.0.0.1', '0.0.0.0'})
            if address[0] in {'::1', '::'}:
                self.assertEqual(len(address), 4)
            else:
                self.assertEqual(len(address), 2)
            for value in address[1:]:
                self.assertIsInstance(value, int)
                self.assertEqual(value, 0)


class MyApp:
    def __init__(self, marker):
        self._marker = marker

    def __call__(self, session, request, bodies):
        return (200, 'OK', {}, self._marker)


def my_build_func(marker):
    return MyApp(marker)


class TestFunctions(TestCase):
    def test_start_server(self):
        marker = os.urandom(16)
        (process, address) = degu._start_server(
            ('127.0.0.1', 0), my_build_func, marker
        )
        self.assertIsInstance(process, multiprocessing.Process)
        self.assertIsInstance(address, tuple)
        self.assertEqual(len(address), 2)
        self.assertEqual(address[0], '127.0.0.1')
        self.assertGreater(address[1], 0)
        client = Client(address)
        conn = client.connect()
        response = conn.request('GET', '/', {}, None)
        self.assertEqual(response.status, 200)
        self.assertEqual(response.reason, 'OK')
        self.assertEqual(response.headers, {'content-length': 16})
        self.assertEqual(response.body.read(), marker)
        conn.close()
        process.terminate()
        process.join()

    def test_start_sslserver(self):
        pki = TempPKI()
        marker = os.urandom(16)
        (process, address) = degu._start_sslserver(
            pki.server_sslconfig, ('127.0.0.1', 0), my_build_func, marker
        )
        self.assertIsInstance(process, multiprocessing.Process)
        self.assertIsInstance(address, tuple)
        self.assertEqual(len(address), 2)
        self.assertEqual(address[0], '127.0.0.1')
        self.assertGreater(address[1], 0)
        client = SSLClient(pki.client_sslconfig, address)
        conn = client.connect()
        response = conn.request('GET', '/', {}, None)
        self.assertEqual(response.status, 200)
        self.assertEqual(response.reason, 'OK')
        self.assertEqual(response.headers, {'content-length': 16})
        self.assertEqual(response.body.read(), marker)
        conn.close()
        process.terminate()
        process.join()


class TestEmbeddedServer(TestCase):
    def test_init(self):
        for address in (degu.IPv4_LOOPBACK, degu.IPv6_LOOPBACK):
            marker = os.urandom(16)
            server = degu.EmbeddedServer(address, my_build_func, marker)
            self.assertIsInstance(server.process, multiprocessing.Process)
            self.assertIsInstance(server.address, tuple)
            self.assertEqual(len(server.address), len(address))
            self.assertEqual(server.address[0], address[0])
            self.assertGreater(server.address[1], 0)
            self.assertEqual(server.build_func, my_build_func)
            self.assertEqual(server.build_args, (marker,))
            self.assertEqual(server.options, {})

            client = Client(server.address)
            conn = client.connect()
            response = conn.request('GET', '/', {}, None)
            self.assertEqual(response.status, 200)
            self.assertEqual(response.reason, 'OK')
            self.assertEqual(response.headers, {'content-length': 16})
            self.assertEqual(response.body.read(), marker)
            conn.close()
            server.terminate()


class TestEmbeddedSSLServer(TestCase):
    def test_start_sslserver(self):
        for address in (degu.IPv4_LOOPBACK, degu.IPv6_LOOPBACK):
            pki = TempPKI()
            marker = os.urandom(16)
            server = degu.EmbeddedSSLServer(
                pki.server_sslconfig, address, my_build_func, marker
            )
            self.assertEqual(server.sslconfig, pki.server_sslconfig)
            self.assertIsInstance(server.process, multiprocessing.Process)
            self.assertIsInstance(server.address, tuple)
            self.assertEqual(len(server.address), len(address))
            self.assertEqual(server.address[0], address[0])
            self.assertGreater(server.address[1], 0)
            self.assertEqual(server.build_func, my_build_func)
            self.assertEqual(server.build_args, (marker,))
            self.assertEqual(server.options, {})

            client = SSLClient(pki.client_sslconfig, server.address)
            conn = client.connect()
            response = conn.request('GET', '/', {}, None)
            self.assertEqual(response.status, 200)
            self.assertEqual(response.reason, 'OK')
            self.assertEqual(response.headers, {'content-length': 16})
            self.assertEqual(response.body.read(), marker)
            conn.close()
            server.terminate()




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for the `degu.util` module`
"""

from unittest import TestCase
from collections import namedtuple

from degu import util


class TestFunctions(TestCase):
    def test_shift_path(self):
        Request = namedtuple('Request', 'mount path')

        # both mount and path are empty:
        request = Request([], [])
        with self.assertRaises(IndexError):
            util.shift_path(request)
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, [])

        # path is empty:
        request = Request(['foo'], [])
        with self.assertRaises(IndexError):
            util.shift_path(request)
        self.assertEqual(request.mount, ['foo'])
        self.assertEqual(request.path, [])

        # start with populated path:
        request = Request([], ['foo', 'bar', 'baz'])
        self.assertEqual(util.shift_path(request), 'foo')
        self.assertEqual(request.mount, ['foo'])
        self.assertEqual(request.path, ['bar', 'baz'])

        self.assertEqual(util.shift_path(request), 'bar')
        self.assertEqual(request.mount, ['foo', 'bar'])
        self.assertEqual(request.path, ['baz'])

        self.assertEqual(util.shift_path(request), 'baz')
        self.assertEqual(request.mount, ['foo', 'bar', 'baz'])
        self.assertEqual(request.path, [])

        with self.assertRaises(IndexError):
            util.shift_path(request)
        self.assertEqual(request.mount, ['foo', 'bar', 'baz'])
        self.assertEqual(request.path, [])

    def test_relative_uri(self):
        Request = namedtuple('Request', 'path query')

        path_permutations = (
            (tuple(),            '/'),
            (('',),              '/'),
            (('foo',),           '/foo'),
            (('foo', ''),        '/foo/'),
            (('foo', 'bar'),     '/foo/bar'),
            (('foo', 'bar', ''), '/foo/bar/'),
        )
        query_permutations = (
            (None,      ''),
            ('',        '?'),
            ('foo',     '?foo'),
            ('foo=bar', '?foo=bar'),
        )
        for (path, uri) in path_permutations:
            request = Request(list(path), None)
            self.assertEqual(util.relative_uri(request), uri)
            self.assertEqual(request.path, list(path))
        for (query, end) in query_permutations:
            request = Request([], query)
            self.assertEqual(util.relative_uri(request), '/' + end)
            self.assertEqual(request.path, [])
        for (path, uri) in path_permutations:
            for (query, end) in query_permutations:
                request = Request(list(path), query)
                self.assertEqual(util.relative_uri(request), uri + end)
                self.assertEqual(request.path, list(path))

        # path is empty:
        request = Request([], None)
        self.assertEqual(util.relative_uri(request), '/')
        self.assertEqual(request.path, [])

        request = Request([], '')
        self.assertEqual(util.relative_uri(request), '/?')
        self.assertEqual(request.path, [])

        request = Request([], 'foo')
        self.assertEqual(util.relative_uri(request), '/?foo')
        self.assertEqual(request.path, [])

        request = Request([], 'foo=bar')
        self.assertEqual(util.relative_uri(request), '/?foo=bar')
        self.assertEqual(request.path, [])

        # No query
        request = Request([''], None)
        self.assertEqual(util.relative_uri(request), '/')
        self.assertEqual(request.path, [''])

        request = Request(['foo'], None)
        self.assertEqual(util.relative_uri(request), '/foo')
        self.assertEqual(request.path, ['foo'])

        request = Request(['foo', ''], None)
        self.assertEqual(util.relative_uri(request), '/foo/')
        self.assertEqual(request.path, ['foo', ''])

        request = Request(['foo', 'bar'], None)
        self.assertEqual(util.relative_uri(request), '/foo/bar')
        self.assertEqual(request.path, ['foo', 'bar'])

        request = Request(['foo', 'bar', ''], None)
        self.assertEqual(util.relative_uri(request), '/foo/bar/')
        self.assertEqual(request.path, ['foo', 'bar', ''])

    def test_absolute_uri(self):
        Request = namedtuple('Request', 'mount path query')

        path_permutations = (
            (tuple(),            '/'),
            (('',),              '/'),
            (('foo',),           '/foo'),
            (('foo', ''),        '/foo/'),
            (('foo', 'bar'),     '/foo/bar'),
            (('foo', 'bar', ''), '/foo/bar/'),
        )
        query_permutations = (
            (None,      ''),
            ('',        '?'),
            ('foo',     '?foo'),
            ('foo=bar', '?foo=bar'),
        )

        for (path, uri) in path_permutations:
            request = Request([], list(path), None)
            self.assertEqual(util.absolute_uri(request), uri)
            self.assertEqual(request.mount, [])
            self.assertEqual(request.path, list(path))
            s = []
            p = list(path)
            while p:
                s.append(p.pop(0))
                request = Request(list(s), list(p), None)
                self.assertEqual(util.absolute_uri(request), uri)
                self.assertEqual(request.mount, s)
                self.assertEqual(request.path, p)

        for (query, end) in query_permutations:
            request = Request([], [], query)
            self.assertEqual(util.absolute_uri(request), '/' + end)
            self.assertEqual(request.mount, [])
            self.assertEqual(request.path, [])

        for (path, uri) in path_permutations:
            for (query, end) in query_permutations:
                request = Request([], list(path), query)
                self.assertEqual(util.absolute_uri(request), uri + end)
                self.assertEqual(request.mount, [])
                self.assertEqual(request.path, list(path))
                s = []
                p = list(path)
                while p:
                    s.append(p.pop(0))
                    request = Request(list(s), list(p), query)
                    self.assertEqual(util.absolute_uri(request), uri + end)
                    self.assertEqual(request.mount, s)
                    self.assertEqual(request.path, p)

        # mount, path, and query are all empty:
        request = Request([], [], None)
        self.assertEqual(util.absolute_uri(request), '/')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, [])

        # only mount:
        request = Request(['foo'], [], None)
        self.assertEqual(util.absolute_uri(request), '/foo')
        self.assertEqual(request.mount, ['foo'])
        self.assertEqual(request.path, [])

        request = Request(['foo', ''], [], None)
        self.assertEqual(util.absolute_uri(request), '/foo/')
        self.assertEqual(request.mount, ['foo', ''])
        self.assertEqual(request.path, [])

        request = Request(['foo', 'bar'], [], None)
        self.assertEqual(util.absolute_uri(request), '/foo/bar')
        self.assertEqual(request.mount, ['foo', 'bar'])
        self.assertEqual(request.path, [])

        request = Request(['foo', 'bar', ''], [], None)
        self.assertEqual(util.absolute_uri(request), '/foo/bar/')
        self.assertEqual(request.mount, ['foo', 'bar', ''])
        self.assertEqual(request.path, [])

        # only path:
        request = Request([], ['foo'], None)
        self.assertEqual(util.absolute_uri(request), '/foo')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, ['foo'])

        request = Request([], ['foo', ''], None)
        self.assertEqual(util.absolute_uri(request), '/foo/')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, ['foo', ''])

        request = Request([], ['foo', 'bar'], None)
        self.assertEqual(util.absolute_uri(request), '/foo/bar')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, ['foo', 'bar'])

        request = Request([], ['foo', 'bar', ''], None)
        self.assertEqual(util.absolute_uri(request), '/foo/bar/')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, ['foo', 'bar', ''])

        # only query:
        request = Request([], [], 'hello')
        self.assertEqual(util.absolute_uri(request), '/?hello')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, [])

        request = Request([], [], 'stuff=junk')
        self.assertEqual(util.absolute_uri(request), '/?stuff=junk')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, [])

        # All of the above:
        request = Request(['foo'], ['bar'], 'hello')
        self.assertEqual(util.absolute_uri(request), '/foo/bar?hello')
        self.assertEqual(request.mount, ['foo'])
        self.assertEqual(request.path, ['bar'])

        request = Request(['foo'], ['bar', ''], 'hello')
        self.assertEqual(util.absolute_uri(request), '/foo/bar/?hello')
        self.assertEqual(request.mount, ['foo'])
        self.assertEqual(request.path, ['bar', ''])

        request = Request(['foo'], ['bar'], 'one=two')
        self.assertEqual(util.absolute_uri(request), '/foo/bar?one=two')
        self.assertEqual(request.mount, ['foo'])
        self.assertEqual(request.path, ['bar'])

        request = Request(['foo'], ['bar', ''], 'one=two')
        self.assertEqual(util.absolute_uri(request), '/foo/bar/?one=two')
        self.assertEqual(request.mount, ['foo'])
        self.assertEqual(request.path, ['bar', ''])




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Some tools for unit testing.

This module imports things that often wouldn't normally be needed except for
unit testing, so thus this separate module helps keep the baseline memory
footprint lower.
"""

import tempfile
import os
import shutil
import multiprocessing
import logging

from .base import _TYPE_ERROR
from .sslhelpers import PKI as _PKI
from .server import Server, SSLServer


log = logging.getLogger(__name__)


class TempPKI(_PKI):
    def __init__(self, client_pki=True, bits=1024):
        # To make unit testing faster, we use 1024 bit keys by default, but this
        # is not the size you should use in production
        ssldir = tempfile.mkdtemp(prefix='TempPKI.')
        super().__init__(ssldir)
        self.server_ca_id = self.create_key(bits)
        self.create_ca(self.server_ca_id)
        self.server_id = self.create_key(bits)
        self.create_csr(self.server_id)
        self.issue_cert(self.server_id, self.server_ca_id)
        if client_pki:
            self.client_ca_id = self.create_key(bits)
            self.create_ca(self.client_ca_id)
            self.client_id = self.create_key(bits)
            self.create_csr(self.client_id)
            self.issue_cert(self.client_id, self.client_ca_id)

    def __del__(self):
        if os.path.isdir(self.ssldir):
            shutil.rmtree(self.ssldir)

    @property
    def server_sslconfig(self):
        return self.get_server_sslconfig(self.server_id, self.client_ca_id)

    @property
    def client_sslconfig(self):
        return self.get_client_sslconfig(self.server_ca_id, self.client_id)

    @property
    def anonymous_server_sslconfig(self):
        return self.get_anonymous_server_sslconfig(self.server_id)

    @property
    def anonymous_client_sslconfig(self):
        return self.get_anonymous_client_sslconfig(self.server_ca_id)


def _run_server(queue, address, app, **options):
    try:
        httpd = Server(address, app, **options)
        queue.put(httpd.address)
        httpd.serve_forever()
    except Exception as e:
        queue.put(e)
        raise e


def _run_sslserver(queue, sslconfig, address, app, **options):
    try:
        httpd = SSLServer(sslconfig, address, app, **options)
        queue.put(httpd.address)
        httpd.serve_forever()
    except Exception as e:
        queue.put(e)
        raise e


def _start_server(address, app, **options):
    import multiprocessing
    queue = multiprocessing.Queue()
    process = multiprocessing.Process(
        target=_run_server,
        args=(queue, address, app),
        kwargs=options,
        daemon=True,
    )
    process.start()
    address = queue.get()
    if isinstance(address, Exception):
        process.terminate()
        process.join()
        raise address
    return (process, address)


def _start_sslserver(sslconfig, address, app, **options):
    if not isinstance(sslconfig, dict):
        raise TypeError(
            _TYPE_ERROR.format('sslconfig', dict, type(sslconfig), sslconfig)
        )
    queue = multiprocessing.Queue()
    process = multiprocessing.Process(
        target=_run_sslserver,
        args=(queue, sslconfig, address, app),
        kwargs=options,
        daemon=True,
    )
    process.start()
    address = queue.get()
    if isinstance(address, Exception):
        process.terminate()
        process.join()
        raise address
    return (process, address)


class _TempProcess:
    def __del__(self):
        self.terminate()

    def terminate(self):
        if getattr(self, 'process', None) is not None:
            self.process.terminate()
            self.process.join()


class TempServer(_TempProcess):
    def __init__(self, address, app, **options):
        (self.process, self.address) = _start_server(address, app, **options)
        self.app = app
        self.options = options

    def __repr__(self):
        return '{}({!r}, {!r})'.format(
            self.__class__.__name__, self.address, self.app
        )


class TempSSLServer(_TempProcess):
    def __init__(self, sslconfig, address, app, **options):
        self.sslconfig = sslconfig
        (self.process, self.address) = _start_sslserver(
            sslconfig, address, app, **options
        )
        self.app = app
        self.options = options

    def __repr__(self):
        return '{}(<sslconfig>, {!r}, {!r})'.format(
            self.__class__.__name__, self.address, self.app
        )


class RequestLogger:
    def __init__(self, app):
        self.app = app

    def __call__(self, session, request, bodies):
        response = self.app(session, request, bodies)
        log.info('[%d] %s %s --> %s %s', session.requests,
            request.method, request.uri, response[0], response[1]
        )
        return response




#!/usr/bin/python3

import timeit

setup = """
gc.enable()

from io import BytesIO

from degu._base import (
    parse_headers,
    parse_content_length,
    parse_range,
    parse_content_range,

    parse_request,
    parse_request_line,
    parse_method,
    parse_uri,

    parse_response,
    parse_response_line,

    format_headers,
    format_request,
    format_response,

    render_headers,
    render_request,
    render_response,

    parse_chunk_size,
    parse_chunk,
)


headers = {
    'content-type': 'application/json',
    'accept': 'application/json',
    'content-length': 12,
    'user-agent': 'Microfiber/14.12.0 (Ubuntu 14.04; x86_64)',
    'x-token': 'VVI5KPPRN5VOG9DITDLEOEIB',
    'extra': 'Super',
    'hello': 'World',
    'k': 'V',
}
headers_src = format_headers(headers).encode()
request = format_request('POST', '/foo/bar?stuff=junk', headers)[:-4]
response = format_response(200, 'OK', headers)[:-4]

dst = memoryview(bytearray(4096))
"""


def run_iter(statement, n):
    for i in range(10):
        t = timeit.Timer(statement, setup)
        yield t.timeit(n)


def run(statement, K=250):
    n = K * 1000
    # Choose fastest of 10 runs:
    elapsed = min(run_iter(statement, n))
    rate = int(n / elapsed)
    print('{:>11,}: {}'.format(rate, statement))
    return rate

def compare(s1, s2, K=100):
    r1 = run(s1, K)
    r2 = run(s2, K)
    percent = (r1 / r2 * 100) - 100
    print('  [{:.1f}%]\n'.format(percent))
    return percent

def test0(s):
    p1 = 'render_headers(dst, '
    p2 = 'format_headers('
    return compare(p1 + s, p2 + s)

def test1(s):
    p1 = 'render_request(dst, '
    p2 = 'format_request('
    return compare(p1 + s, p2 + s)

def test2(s, K=250):
    p1 = 'render_response(dst, '
    p2 = 'format_response('
    return compare(p1 + s, p2 + s)


print('-' * 80)

test0('headers)')
test0("{'content-length': 17})")
test0("{'content-length': 17, 'content-type': 'text/plain'})")
print('')
test1("'GET', '/foo', {})")
test1("'PUT', '/foo', {'content-length': 17})")
test1("'PUT', '/foo', {'content-length': 17, 'content-type': 'text/plain'})")
test1("'PUT', '/foo', headers)")
print('')
test2("200, 'OK', {})")
test2("200, 'OK', {'content-length': 17})")
test2("200, 'OK', {'content-length': 17, 'content-type': 'text/plain'})")
test2("200, 'OK', headers)")

print('\nHeader parsing:')
run('parse_headers(headers_src)')
run("parse_headers(b'Content-Length: 123456')")
run("parse_headers(b'Transfer-Encoding: chunked')")
run("parse_headers(b'Content-Length: 123456\\r\\nContent-Type: application/json')")
run("parse_headers(b'Transfer-Encoding: chunked\\r\\nContent-Type: application/json')")
run("parse_headers(b'Range: bytes=0-0')")
run("parse_headers(b'Range: bytes=9999999999999998-9999999999999998')")
run("parse_content_length(b'0')")
run("parse_content_length(b'9999999999999999')")
run("parse_range(b'bytes=0-0')")
run("parse_range(b'bytes=9999999999999998-9999999999999998')")
run("parse_content_range(b'bytes 0-0/1')")
run("parse_content_range(b'bytes 9999999999999998-9999999999999998/9999999999999999')")

print('\nRequest parsing:')
run('parse_request(request, BytesIO())')
run("parse_request(b'GET / HTTP/1.1', BytesIO())")
run("parse_request(b'DELETE /foo/bar?stuff=junk HTTP/1.1', BytesIO())")
run("parse_request(b'GET / HTTP/1.1\\r\\ncontent-length: 17', BytesIO())")
run("parse_request_line(b'GET / HTTP/1.1')")
run("parse_request_line(b'DELETE /foo/bar?stuff=junk HTTP/1.1')")
run("parse_method(b'GET')")
run("parse_method(b'PUT')")
run("parse_method(b'POST')")
run("parse_method(b'HEAD')")
run("parse_method(b'DELETE')")
run("parse_uri(b'/')")
run("parse_uri(b'/?')")
run("parse_uri(b'/foo/bar')")
run("parse_uri(b'/foo/bar?stuff=junk')")

print('\nResponse parsing:')
run("parse_response('GET', response, BytesIO())")
run("parse_response_line(b'HTTP/1.1 200 OK')")
run("parse_response_line(b'HTTP/1.1 404 Not Found')")

print('\nChunk parsing:')
run("parse_chunk_size(b'0')")
run("parse_chunk_size(b'1000000')")
run("parse_chunk(b'0')")
run("parse_chunk(b'1000000')")
run("parse_chunk(b'0;key=value')")
run("parse_chunk(b'1000000;key=value')")

print('\nHeader formating:')
run('render_headers(dst, headers)')
run('format_headers(headers)')
run('format_headers({})')
run("format_headers({'content-length': 17})")
run("format_headers({'content-length': 17, 'content-type': 'text/plain'})")

print('\nRequest formatting:')
run("format_request('GET', '/foo', {})")
run("format_request('PUT', '/foo', {'content-length': 17})")
run("format_request('PUT', '/foo', headers)")
run("render_request(dst, 'PUT', '/foo', headers)")

print('\nResponse formatting:')
run("format_response(200, 'OK', {})")
run("format_response(200, 'OK', {'content-length': 17})")
run("format_response(200, 'OK', headers)")
run("render_response(dst, 200, 'OK', headers)")


print('-' * 80)




# degu: an embedded HTTP server and client library
# Copyright (C) 2014-2016 Novacut Inc
#
# This file is part of `degu`.
#
# `degu` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `degu` is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with `degu`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Unit tests for the `degu.rgi` module`
"""

from unittest import TestCase
import os
import string
from random import SystemRandom
from copy import deepcopy
from collections import namedtuple

from degu import rgi


random = SystemRandom()


def random_identifier():
    return ''.join(random.choice(string.ascii_lowercase) for i in range(17))


def random_value():
    return os.urandom(10)


class MockObject:
    def __init__(self, **kw):
        for (key, value) in kw.items():
            assert not key.startswith('_')
            setattr(self, key, value)


class Session(MockObject):
    """
    Mock class used for RGI *session* argument.
    """

class Body(MockObject):
    """
    Mock class used for bodies.Body.
    """


class BodyIter(MockObject):
    """
    Mock class used for bodies.BodyIter.
    """


class ChunkedBody(MockObject):
    """
    Mock class used for bodies.ChunkedBody.
    """


class ChunkedBodyIter(MockObject):
    """
    Mock class used for bodies.ChunkedBodyIter.
    """


Bodies = namedtuple('Bodies', 'Body BodyIter ChunkedBody ChunkedBodyIter')
default_bodies = Bodies(Body, BodyIter, ChunkedBody, ChunkedBodyIter)


def build_session(**kw):
    base = {
        'address': ('127.0.0.1', 52521),
        'requests': 0,
        'store': {},
    }
    base.update(kw)
    return Session(**base)


def build_request(**kw):
    request = {
        'method': 'POST',
        'uri': '/foo/bar?stuff=junk',
        'mount': ['foo'],
        'path': ['bar'],
        'query': 'stuff=junk',
        'headers': {},
        'body': None,
    }
    for (key, value) in kw.items():
        request[key] = value
    return MockObject(**request)


class TestMockObject(TestCase):
    def test_init(self):
        for klass in (Body, BodyIter, ChunkedBody, ChunkedBodyIter):
            # No kw args:
            body = klass()
            self.assertIsInstance(body, MockObject)
            self.assertEqual(
                list(filter(lambda n: not n.startswith('_'), dir(body))),
                []
            )

            # One kw arg:
            key1 = random_identifier()
            val1 = random_value()
            kw = {key1: val1}
            body = klass(**kw)
            self.assertIsInstance(body, MockObject)
            self.assertEqual(
                list(filter(lambda n: not n.startswith('_'), dir(body))),
                [key1]
            )
            self.assertIs(getattr(body, key1), val1)

            # Two kw args:
            key2 = random_identifier()
            val2 = random_value()
            kw = {key1: val1, key2: val2}
            body = klass(**kw)
            self.assertIsInstance(body, MockObject)
            self.assertEqual(
                list(filter(lambda n: not n.startswith('_'), dir(body))),
                sorted([key1, key2])
            )
            self.assertIs(getattr(body, key1), val1)
            self.assertIs(getattr(body, key2), val2)

            # Three kw args:
            key3 = random_identifier()
            val3 = random_value()
            kw = {key1: val1, key2: val2, key3: val3}
            body = klass(**kw)
            self.assertIsInstance(body, MockObject)
            self.assertEqual(
                list(filter(lambda n: not n.startswith('_'), dir(body))),
                sorted([key1, key2, key3])
            )
            self.assertIs(getattr(body, key1), val1)
            self.assertIs(getattr(body, key2), val2)
            self.assertIs(getattr(body, key3), val3)


class TestFunctions(TestCase):
    def test_getattr(self):
        class Example:
            def __init__(self, key, value):
                setattr(self, key, value)

        label = random_identifier()
        key = random_identifier()
        value = random_value()
        obj = Example(key, value)

        # Attribute is present:
        (L, V) = rgi._getattr(label, obj, key)
        self.assertEqual(L, '{}.{}'.format(label, key))
        self.assertIs(V, value)

        # Attribute is missing:
        key2 = random_identifier()
        with self.assertRaises(ValueError) as cm:
            rgi._getattr(label, obj, key2)
        self.assertEqual(str(cm.exception),
            "{}: 'Example' object has no attribute {!r}".format(label, key2)
        )

    def test_ensure_attr_is(self):
        class Example:
            def __init__(self, key, value):
                setattr(self, key, value)

        label = random_identifier()
        key = random_identifier()
        value = random_value()
        obj = Example(key, value)

        # Attribute is expected:
        self.assertIsNone(rgi._ensure_attr_is(label, obj, key, value))

        # Attribute is not expected:
        value2 = random_value()
        with self.assertRaises(ValueError) as cm:
            rgi._ensure_attr_is(label, obj, key, value2)
        self.assertEqual(str(cm.exception),
            "{}.{} must be {!r}; got {!r}".format(label, key, value2, value)
        )

        # Attribute is missing:
        key2 = random_identifier()
        with self.assertRaises(ValueError) as cm:
            rgi._ensure_attr_is(label, obj, key2, value)
        self.assertEqual(str(cm.exception),
            "{}: 'Example' object has no attribute {!r}".format(label, key2)
        )

    def test_check_dict(self):
        # obj is not a dict:
        label = random_identifier()
        obj = random_identifier()
        with self.assertRaises(TypeError) as cm:
            rgi._check_dict(label, obj)
        self.assertEqual(str(cm.exception),
            rgi.TYPE_ERROR.format(label, dict, str, obj)
        )

        # obj contains a non-string key:
        obj = dict(
            (random_identifier(), random_value()) for i in range(5)
        )
        key = random_identifier().encode()
        obj[key] = random_value()
        with self.assertRaises(TypeError) as cm:
            rgi._check_dict(label, obj)
        self.assertEqual(str(cm.exception),
            '{}: keys must be {!r}; got a {!r}: {!r}'.format(label, str, bytes, key)
        )

        # All good:
        obj = dict(
            (random_identifier(), random_value()) for i in range(6)
        )
        self.assertIsNone(rgi._check_dict(label, obj))

    def test_check_headers(self):
        # headers is not a dict:
        label = random_identifier()
        headers = random_identifier()
        with self.assertRaises(TypeError) as cm:
            rgi._check_headers(label, headers)
        self.assertEqual(str(cm.exception),
            rgi.TYPE_ERROR.format(label, dict, str, headers)
        )

        # headers contains a non-string key:
        headers = dict(
            (random_identifier(), random_identifier()) for i in range(5)
        )
        key = random_identifier().encode()
        headers[key] = random_identifier()
        with self.assertRaises(TypeError) as cm:
            rgi._check_headers(label, headers)
        self.assertEqual(str(cm.exception),
            '{}: keys must be {!r}; got a {!r}: {!r}'.format(label, str, bytes, key)
        )

        # headers contains a non-casefolded key:
        headers = dict(
            (random_identifier(), random_identifier()) for i in range(5)
        )
        key = random_identifier().upper()
        headers[key] = random_identifier()
        with self.assertRaises(ValueError) as cm:
            rgi._check_headers(label, headers)
        self.assertEqual(str(cm.exception),
            '{}: non-casefolded header name: {!r}'.format(label, key)
        )

        # headers contains a non-string value:
        headers = dict(
            (random_identifier(), random_identifier()) for i in range(5)
        )
        key = random_identifier()
        value = random_value()
        headers[key] = value
        with self.assertRaises(TypeError) as cm:
            rgi._check_headers(label, headers)
        self.assertEqual(str(cm.exception),
            '{}[{!r}]: need a {!r}; got a {!r}: {!r}'.format(label, key, str, bytes, value)
        )

        # content-length plus tranfer-encoding
        headers = dict(
            (random_identifier(), random_identifier()) for i in range(5)
        )
        headers['content-length'] = random_identifier()
        headers['transfer-encoding'] = random_identifier()
        with self.assertRaises(ValueError) as cm:
            rgi._check_headers(label, headers)
        self.assertEqual(str(cm.exception),
            '{}: content-length and transfer-encoding in headers'.format(label)
        )

        # content-length isn't an int:
        headers = dict(
            (random_identifier(), random_identifier()) for i in range(5)
        )
        headers['content-length'] = '17'
        with self.assertRaises(TypeError) as cm:
            rgi._check_headers(label, headers)
        self.assertEqual(str(cm.exception),
            "{}['content-length']: need a {!r}; got a {!r}: '17'".format(label, int, str)
        )

        # content-length is negative:
        headers = dict(
            (random_identifier(), random_identifier()) for i in range(5)
        )
        headers['content-length'] = -1
        with self.assertRaises(ValueError) as cm:
            rgi._check_headers(label, headers)
        self.assertEqual(str(cm.exception),
            "{}['content-length']: must be >=0; got -1".format(label)
        )

        # Bad transfer-encoding:
        headers = dict(
            (random_identifier(), random_identifier()) for i in range(5)
        )
        headers['transfer-encoding'] = 'clumped'
        with self.assertRaises(ValueError) as cm:
            rgi._check_headers(label, headers)
        self.assertEqual(str(cm.exception),
            "{}['transfer-encoding']: must be 'chunked'; got 'clumped'".format(label)
        )

        # All good:
        label = random_identifier()
        headers = dict(
            (random_identifier(), random_identifier()) for i in range(6)
        )
        self.assertIsNone(rgi._check_headers(label, headers))

        # All good, with a content-length:
        label = random_identifier()
        headers = dict(
            (random_identifier(), random_identifier()) for i in range(5)
        )
        headers['content-length'] = 0
        self.assertIsNone(rgi._check_headers(label, headers))

        # All good, with a transfer-encoding:
        label = random_identifier()
        headers = dict(
            (random_identifier(), random_identifier()) for i in range(5)
        )
        headers['transfer-encoding'] = 'chunked'
        self.assertIsNone(rgi._check_headers(label, headers))

    def test_check_address(self):
        # Bad address type:
        label = random_identifier()
        address = ['127.0.0.1', 12345]
        with self.assertRaises(TypeError) as cm:
            rgi._check_address(label, address)
        self.assertEqual(str(cm.exception),
            rgi.TYPE_ERROR.format(label, (tuple, str, bytes), list, address)
        )
        label = random_identifier()
        address = ['::1', 12345, 0, 0]
        with self.assertRaises(TypeError) as cm:
            rgi._check_address(label, address)
        self.assertEqual(str(cm.exception),
            rgi.TYPE_ERROR.format(label, (tuple, str, bytes), list, address)
        )

        # Tuple of wrong length:
        label = random_identifier()
        address = ('::1', 12345, 0)
        with self.assertRaises(ValueError) as cm:
            rgi._check_address(label, address)
        self.assertEqual(str(cm.exception),
            '{}: tuple must have 2 or 4 items; got {!r}'.format(label, address)
        )
        label = random_identifier()
        address = ('::1', 12345, 0, 0, 0)
        with self.assertRaises(ValueError) as cm:
            rgi._check_address(label, address)
        self.assertEqual(str(cm.exception),
            '{}: tuple must have 2 or 4 items; got {!r}'.format(label, address)
        )

        # Test all 4 types of good values:
        good = (
            ('127.0.0.1', 12345),
            ('::1', 23456, 0, 0),
            '/tmp/random/my.socket',
            b'\x0000022',
        )
        for address in good:    
            label = random_identifier()
            self.assertIsNone(rgi._check_address(label, address))

    def test_validate_session(self):
        # Missing required attributes:
        good = {
            'address': ('127.0.0.1', 52521),
            'requests': 0,
            'store': {},
        }
        sess = Session(**good)
        self.assertIsNone(rgi._validate_session(sess))
        for key in sorted(good):
            bad = deepcopy(good)
            del bad[key]
            sess = Session(**bad)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_session(sess)
            self.assertEqual(str(cm.exception),
                "session: 'Session' object has no attribute {!r}".format(key)
            )

        # Bad session.address type:
        address = ['127.0.0.1', 12345]
        bad = deepcopy(good)
        bad['address'] = address
        sess = Session(**bad)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_session(sess)
        self.assertEqual(str(cm.exception),
            rgi.TYPE_ERROR.format('session.address', (tuple, str, bytes), list, address)
        )

        # session.address tuple is wrong length:
        address = ('::1', 2345, 0, 0, 0)
        bad = deepcopy(good)
        bad['address'] = address
        with self.assertRaises(ValueError) as cm:
            rgi._validate_session(Session(**bad))
        self.assertEqual(str(cm.exception),
            "session.address: tuple must have 2 or 4 items; got ('::1', 2345, 0, 0, 0)"
        )

        # session.requests isn't an `int`:
        bad = deepcopy(good)
        bad['requests'] = 0.0
        with self.assertRaises(TypeError) as cm:
            rgi._validate_session(Session(**bad))
        self.assertEqual(str(cm.exception),
            "session.requests: need a <class 'int'>; got a <class 'float'>: 0.0"
        )

        # session.requests is negative:
        bad = deepcopy(good)
        bad['requests'] = -1
        with self.assertRaises(ValueError) as cm:
            rgi._validate_session(Session(**bad))
        self.assertEqual(str(cm.exception),
            'session.requests must be >= 0; got -1'
        )

    def test_reconstruct_uri(self):
        # mount, path, and query are all empty:
        kw = {'mount': [], 'path': [], 'query': None}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, [])

        # only mount:
        kw = {'mount': ['foo'], 'path': [], 'query': None}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/foo')
        self.assertEqual(request.mount, ['foo'])
        self.assertEqual(request.path, [])

        kw = {'mount': ['foo', ''], 'path': [], 'query': None}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/foo/')
        self.assertEqual(request.mount, ['foo', ''])
        self.assertEqual(request.path, [])

        kw = {'mount': ['foo', 'bar'], 'path': [], 'query': None}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/foo/bar')
        self.assertEqual(request.mount, ['foo', 'bar'])
        self.assertEqual(request.path, [])

        kw = {'mount': ['foo', 'bar', ''], 'path': [], 'query': None}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/foo/bar/')
        self.assertEqual(request.mount, ['foo', 'bar', ''])
        self.assertEqual(request.path, [])

        # only path:
        kw = {'mount': [], 'path': ['foo'], 'query': None}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/foo')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, ['foo'])

        kw = {'mount': [], 'path': ['foo', ''], 'query': None}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/foo/')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, ['foo', ''])

        kw = {'mount': [], 'path': ['foo', 'bar'], 'query': None}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/foo/bar')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, ['foo', 'bar'])

        kw = {'mount': [], 'path': ['foo', 'bar', ''], 'query': None}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/foo/bar/')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, ['foo', 'bar', ''])

        # only query:
        kw = {'mount': [], 'path': [], 'query': 'hello'}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/?hello')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, [])

        kw = {'mount': [], 'path': [], 'query': 'stuff=junk'}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/?stuff=junk')
        self.assertEqual(request.mount, [])
        self.assertEqual(request.path, [])

        # All of the above:
        kw = {'mount': ['foo'], 'path': ['bar'], 'query': 'hello'}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/foo/bar?hello')
        self.assertEqual(request.mount, ['foo'])
        self.assertEqual(request.path, ['bar'])

        kw = {'mount': ['foo'], 'path': ['bar', ''], 'query': 'hello'}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/foo/bar/?hello')
        self.assertEqual(request.mount, ['foo'])
        self.assertEqual(request.path, ['bar', ''])

        kw = {'mount': ['foo'], 'path': ['bar'], 'query': 'one=two'}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/foo/bar?one=two')
        self.assertEqual(request.mount, ['foo'])
        self.assertEqual(request.path, ['bar'])

        kw = {'mount': ['foo'], 'path': ['bar', ''], 'query': 'one=two'}
        request = MockObject(**kw)
        self.assertEqual(rgi._reconstruct_uri(request), '/foo/bar/?one=two')
        self.assertEqual(request.mount, ['foo'])
        self.assertEqual(request.path, ['bar', ''])

    def test_check_uri_invariant(self):
        good = {
            'uri': '/foo/bar/baz?stuff=junk',
            'mount': [],
            'path': ['foo', 'bar', 'baz'],
            'query': 'stuff=junk',
        }
        request = MockObject(**good)
        self.assertIsNone(rgi._check_uri_invariant(request))

        bad = {
            'uri': '/foo/bar/baz?stuff=junk',
            'mount': ['foo'],
            'path': ['baz'],
            'query': 'stuff=junk',
        }
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._check_uri_invariant(request)
        self.assertEqual(str(cm.exception),
            "reconstruct_uri(request) != request.uri: "
            "'/foo/baz?stuff=junk' != '/foo/bar/baz?stuff=junk'"
        )

    def test_validate_request(self):
        good = {
            'method': 'POST',
            'uri': '/foo/bar?stuff=junk',
            'mount': ['foo'],
            'path': ['bar'],
            'query': 'stuff=junk',
            'headers': {},
            'body': None,
        }
        request = MockObject(**good)
        self.assertIsNone(rgi._validate_request(default_bodies, request))
        for key in sorted(good):
            bad = deepcopy(good)
            del bad[key]
            request = MockObject(**bad)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_request(default_bodies, request)
            self.assertEqual(str(cm.exception),
                "request: 'MockObject' object has no attribute {!r}".format(key)
            )

        # Bad request.method value:
        bad = deepcopy(good)
        bad['method'] = 'OPTIONS'
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.method: value 'OPTIONS' not in ('GET', 'PUT', 'POST', 'DELETE', 'HEAD')"
        )

        # Bad request.uri type:
        bad = deepcopy(good)
        bad['uri'] = bad['uri'].encode()
        request = MockObject(**bad)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.uri: need a <class 'str'>; got a <class 'bytes'>: b'/foo/bar?stuff=junk'"
        )

        # Bad request.mount type:
        bad = deepcopy(good)
        bad['mount'] = ('foo',)
        request = MockObject(**bad)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.mount: need a <class 'list'>; got a <class 'tuple'>: ('foo',)"
        )

        # Bad request.mount[0] type:
        bad = deepcopy(good)
        bad['mount'] = [b'foo']
        request = MockObject(**bad)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.mount[0]: need a <class 'str'>; got a <class 'bytes'>: b'foo'"
        )

        # Bad request.mount[1] type:
        bad = deepcopy(good)
        bad['mount'] = ['foo', b'baz']
        request = MockObject(**bad)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.mount[1]: need a <class 'str'>; got a <class 'bytes'>: b'baz'"
        )

        # Bad request.path type:
        bad = deepcopy(good)
        bad['path'] = ('bar',)
        request = MockObject(**bad)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.path: need a <class 'list'>; got a <class 'tuple'>: ('bar',)"
        )

        # Bad request.path[0] type:
        bad = deepcopy(good)
        bad['path'] = [b'bar']
        request = MockObject(**bad)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.path[0]: need a <class 'str'>; got a <class 'bytes'>: b'bar'"
        )

        # Bad request.path[1] type:
        bad = deepcopy(good)
        bad['path'] = ['bar', b'baz']
        request = MockObject(**bad)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.path[1]: need a <class 'str'>; got a <class 'bytes'>: b'baz'"
        )

        # Bad request.query type:
        bad = deepcopy(good)
        bad['query'] = {'stuff': 'junk'}
        request = MockObject(**bad)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.query: need a <class 'str'>; got a <class 'dict'>: {'stuff': 'junk'}"
        )

        # Bad request.headers type:
        bad = deepcopy(good)
        bad['headers'] = [('content-length', 17)]
        request = MockObject(**bad)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.headers: need a <class 'dict'>; got a <class 'list'>: [('content-length', 17)]"
        )

        # Bad request.body type:
        bad_bodies = (BodyIter(), ChunkedBodyIter())
        body_types = (Body, ChunkedBody)
        for body in bad_bodies:
            bad = deepcopy(good)
            bad['body'] = body
            request = MockObject(**bad)
            with self.assertRaises(TypeError) as cm:
                rgi._validate_request(default_bodies, request)
            self.assertEqual(str(cm.exception),
                rgi.TYPE_ERROR.format(
                    'request.body', body_types, type(body), body
                )
            )

        ############################
        # request.body is `None`:

        # 'content-length' header is present:
        bad = deepcopy(good)
        bad['headers']['content-length'] = 17
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body is None but 'content-length' header is included"
        )

        # 'transfer-encoding' header is present:
        bad = deepcopy(good)
        bad['headers']['transfer-encoding'] = 'chunked'
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body is None but 'transfer-encoding' header is included"
        )

        # All valid request.method should work when request.body is None:
        for method in rgi.REQUEST_METHODS:
            also_good = deepcopy(good)
            also_good['method'] = method
            request = MockObject(**also_good)
            self.assertIsNone(rgi._validate_request(default_bodies, request))

        #######################################
        # request.body is a `Body` instance:

        # Body is missing 'chunked' attribute:
        bad = deepcopy(good)
        bad['body'] = Body(content_length=17, state=0)
        bad['headers']['content-length'] = 17
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body: 'Body' object has no attribute 'chunked'"
        )

        # Body.chunked is True:
        bad = deepcopy(good)
        bad['body'] = Body(chunked=True)
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body.chunked must be False; got True"
        )

        # Body with 'transfer-encoding' header:
        bad = deepcopy(good)
        bad['body'] = Body(chunked=False)
        bad['headers']['transfer-encoding'] = 'chunked'
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body: bodies.Body with 'transfer-encoding' header"
        )

        # Body.content_length attribute is missing:
        bad = deepcopy(good)
        bad['body'] = Body(chunked=False)
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body: 'Body' object has no attribute 'content_length'"
        )

        # Body without 'content-length' header:
        bad = deepcopy(good)
        bad['body'] = Body(chunked=False, content_length=17)
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body: bodies.Body, but missing 'content-length' header"
        )

        # Body.content_length != headers['content-length']:
        bad = deepcopy(good)
        bad['body'] = Body(chunked=False, content_length=17)
        bad['headers']['content-length'] = 16
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body.content_length != request.headers['content-length']: 17 != 16"
        )

        # Body missing 'state' attribute:
        bad = deepcopy(good)
        bad['body'] = Body(chunked=False, content_length=17)
        bad['headers']['content-length'] = 17
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body: 'Body' object has no attribute 'state'"
        )

        # Body.states must be False prior to calling the application:
        bad = deepcopy(good)
        bad['body'] = Body(chunked=False, content_length=17, state=2)
        bad['headers']['content-length'] = 17
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body.state must equal 0; got 2"
        )

        # Methods that should work when request.body is a Body instance:
        for method in ('PUT', 'POST'):
            also_good = deepcopy(good)
            also_good['method'] = method
            also_good['body'] = Body(state=0, chunked=False, content_length=17)
            also_good['headers']['content-length'] = 17
            request = MockObject(**also_good)
            self.assertIsNone(rgi._validate_request(default_bodies, request))

        # Methods that should *not* work when request.body is a Body instance:
        for method in ('GET', 'HEAD', 'DELETE'):
            bad = deepcopy(good)
            bad['method'] = method
            bad['body'] = Body(state=0, chunked=False, content_length=17)
            bad['headers']['content-length'] = 17
            request = MockObject(**bad)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_request(default_bodies, request)
            self.assertEqual(str(cm.exception),
                "request.method cannot be {!r} when request.body is not None".format(method)
            )

        ##############################################
        # request.body is a `ChunkedBody` instance:

        # ChunkedBody is missing 'chunked' attribute:
        bad = deepcopy(good)
        bad['body'] = ChunkedBody(state=0)
        bad['headers']['transfer-encoding'] = 'chunked'
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body: 'ChunkedBody' object has no attribute 'chunked'"
        )

        # ChunkedBody.chunked is False:
        bad = deepcopy(good)
        bad['body'] = ChunkedBody(state=0, chunked=False)
        bad['headers']['transfer-encoding'] = 'chunked'
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body.chunked must be True; got False"
        )

        # ChunkedBody with 'content-length' header:
        bad = deepcopy(good)
        bad['body'] = ChunkedBody(chunked=True, state=0)
        bad['headers']['content-length'] = 17
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body: bodies.ChunkedBody with 'content-length' header"
        )

        # ChunkedBody without 'transfer-encoding' header:
        bad = deepcopy(good)
        bad['body'] = ChunkedBody(chunked=True, state=0)
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body: bodies.ChunkedBody, but missing 'transfer-encoding' header"
        )

        # ChunkedBody is missing 'state' attribute:
        bad = deepcopy(good)
        bad['body'] = ChunkedBody(chunked=True)
        bad['headers']['transfer-encoding'] = 'chunked'
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            "request.body: 'ChunkedBody' object has no attribute 'state'"
        )

        # ChunkedBody.state must be 0 prior to calling the application:
        bad = deepcopy(good)
        bad['body'] = ChunkedBody(chunked=True, state=2)
        bad['headers']['transfer-encoding'] = 'chunked'
        request = MockObject(**bad)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_request(default_bodies, request)
        self.assertEqual(str(cm.exception),
            'request.body.state must equal 0; got 2'
        )

        # Methods that should work when request.body is a ChunkedBody instance:
        for method in ('PUT', 'POST'):
            bad = deepcopy(good)
            bad['method'] = method
            bad['body'] = ChunkedBody(state=0, chunked=True)
            bad['headers']['transfer-encoding'] = 'chunked'
            request = MockObject(**bad)
            self.assertIsNone(rgi._validate_request(default_bodies, request))

        # Methods that should *not* work when request.body is a ChunkedBody instance:
        for method in ('GET', 'HEAD', 'DELETE'):
            bad = deepcopy(good)
            bad['method'] = method
            bad['body'] = ChunkedBody(state=0, chunked=True)
            bad['headers']['transfer-encoding'] = 'chunked'
            request = MockObject(**bad)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_request(default_bodies, request)
            self.assertEqual(str(cm.exception),
                "request.method cannot be {!r} when request.body is not None".format(method)
            )

    def test_validate_response(self):
        # Validator.__call__() will pass in the *bodies* argument, by which the
        # the Body, BodyIter, ChunkedBody, and ChunkedBodyIter classes are
        # exposed in a server-agnostic fashion.

        request = MockObject(method='GET')

        # response isn't a `tuple`:
        bad = [200, 'OK', {}, None]
        with self.assertRaises(TypeError) as cm:
            rgi._validate_response(default_bodies, deepcopy(request), bad)
        self.assertEqual(str(cm.exception),
            rgi.TYPE_ERROR.format('response', tuple, list, bad)
        )

        # len(response) != 4:
        bad = (200, 'OK', {})
        with self.assertRaises(ValueError) as cm:
            rgi._validate_response(default_bodies, deepcopy(request), bad)
        self.assertEqual(str(cm.exception), 'len(response) must be 4, got 3')
        bad = (200, 'OK', {}, None, None)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_response(default_bodies, deepcopy(request), bad)
        self.assertEqual(str(cm.exception), 'len(response) must be 4, got 5')

        # response status isn't an int:
        bad = ('200', 'OK', {}, None)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_response(default_bodies, deepcopy(request), bad)
        self.assertEqual(str(cm.exception),
            rgi.TYPE_ERROR.format('response[0]', int, str, '200')
        )

        # response status < 100:
        bad = (99, 'OK', {}, None)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_response(default_bodies, deepcopy(request), bad)
        self.assertEqual(str(cm.exception),
            'response[0]: need 100 <= status <= 599; got 99'
        )

        # response status > 599:
        bad = (600, 'OK', {}, None)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_response(default_bodies, deepcopy(request), bad)
        self.assertEqual(str(cm.exception),
            'response[0]: need 100 <= status <= 599; got 600'
        )

        # Test all valid response status:
        for status in range(100, 600):
            self.assertTrue(100 <= status <= 599)
            good = (status, 'OK', {}, None)
            self.assertIsNone(
                rgi._validate_response(default_bodies, deepcopy(request), good)
            )

        # response reason isn't an str:
        bad = (200, b'OK', {}, None)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_response(default_bodies, deepcopy(request), bad)
        self.assertEqual(str(cm.exception),
            rgi.TYPE_ERROR.format('response[1]', str, bytes, b'OK')
        )

        # response reason is empty:
        bad = (200, '', {}, None)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_response(default_bodies, deepcopy(request), bad)
        self.assertEqual(str(cm.exception),
            "response[1]: reason cannot be an empty ''"
        )

        # response headers isn't a dict:
        bad = (200, 'OK', [('foo', 'BAR')], None)
        with self.assertRaises(TypeError) as cm:
            rgi._validate_response(default_bodies, deepcopy(request), bad)
        self.assertEqual(str(cm.exception),
            rgi.TYPE_ERROR.format('response[2]', dict, list, [('foo', 'BAR')])
        )

        # request method is 'HEAD', but response headers include neither
        # 'content-length' nor 'transfer-encoding'
        r = MockObject(method='HEAD')
        bad = (200, 'OK', {}, None)
        with self.assertRaises(ValueError) as cm:
            rgi._validate_response(default_bodies, r, bad)
        self.assertEqual(str(cm.exception),
            "response[2]: response to HEAD request must include 'content-length' or 'transfer-encoding' header"
        )

        # Test valid responses to 'HEAD' request:
        r = MockObject(method='HEAD')
        good = (200, 'OK', {'content-length': 17}, None)
        self.assertIsNone(rgi._validate_response(default_bodies, r, good))
        good = (200, 'OK', {'transfer-encoding': 'chunked'}, None)
        self.assertIsNone(rgi._validate_response(default_bodies, r, good))

        # bad response body type:
        bad = (200, 'OK', {}, 'hello')
        with self.assertRaises(TypeError) as cm:
            rgi._validate_response(default_bodies, deepcopy(request), bad)
        self.assertEqual(str(cm.exception),
            'response[3]: bad response body type: {!r}'.format(str)
        )

        # 'HEAD' request, but response body isn't None:
        bad_bodies = (
            b'D' * 17,
            bytearray(b'D' * 17),
            Body(state=0, chunked=False, content_length=17),
            BodyIter(state=0, chunked=False, content_length=17),
            ChunkedBody(state=0, chunked=False),
            ChunkedBodyIter(state=0, chunked=False),
        )
        for body in bad_bodies:
            r = MockObject(method='HEAD')
            if isinstance(body, (ChunkedBody, ChunkedBodyIter)):
                headers = {'transfer-encoding': 'chunked'}
            else:
                headers = {'content-length': 17}
            bad = (200, 'OK', headers, body)
            with self.assertRaises(TypeError) as cm:
                rgi._validate_response(default_bodies, r, bad)
            self.assertEqual(str(cm.exception),
                "response[3]: must be None when request.method is 'HEAD'; got a {!r}".format(
                    type(body)
                )
            )

        # response body is None, but 'content-length' or 'transfer-encoding'
        # header is included:
        bad_headers = (
            ('content-length', 17),
            ('transfer-encoding', 'chunked'),
        )
        for (k, v) in bad_headers:
            bad = (200, 'OK', {k: v}, None)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                '{}: response body is None, but {!r} header is included'.format(
                    'response[3]', k
                )
            )

        # response body is (bytes, bytarray, Body, BodyIter), but
        # 'transfer-encoding' header included:
        bad_bodies = (
            b'hello',
            bytearray(b'hello'),
            Body(content_length=5, chunked=False, state=2),
            BodyIter(content_length=5, chunked=False, state=2),
        )
        for body in bad_bodies:
            bad = (200, 'OK', {'transfer-encoding': 'chunked'}, body)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                '{}: response body is {!r}, but {!r} header is included'.format(
                    'response[3]', type(body), 'transfer-encoding'
                )
            )

        # length mismatch when body is (bytes, bytearray):
        for body in (b'hello', bytearray(b'hello')):
            bad = (200, 'OK', {'content-length': 17}, body)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                "response[3]: len(body) is 5, but 'content-length' is 17"
            )

        # response body is (bytes, bytearray), but 'content-length' header is 
        # missing (should be filled in by server before writing response):
        for body in (b'hello', bytearray(b'hello')):
            good = (200, 'OK', {}, body)
            self.assertIsNone(
                rgi._validate_response(default_bodies, deepcopy(request), good)
            )

        # response body is (Body, BodyIter), but missing 'content_length'
        # attribute:
        for klass in (Body, BodyIter):
            body = klass(chunked=False, state=0)
            bad = (200, 'OK', {'content-length': 17}, body)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                "response[3]: {!r} object has no attribute 'content_length'".format(
                    klass.__name__
                )
            )

        # length mismatch when body is (Body, BodyIter):
        for klass in (Body, BodyIter):
            body = klass(chunked=False, state=0, content_length=18)
            bad = (200, 'OK', {'content-length': 17}, body)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                "response[3]: body.content_length is 18, but 'content-length' is 17"
            )

        # response body is (Body, BodyIter), but 'content-length' header is 
        # missing (should be filled in by server before writing response):
        for klass in (Body, BodyIter):
            body = klass(chunked=False, state=0, content_length=18)
            good = (200, 'OK', {}, body)
            self.assertIsNone(
                rgi._validate_response(default_bodies, deepcopy(request), good)
            )

        # response body is (Body, BodyIter), but missing 'chunked' attribute:
        for klass in (Body, BodyIter):
            body = klass(state=0, content_length=17)
            bad = (200, 'OK', {'content-length': 17}, body)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                "response[3]: {!r} object has no attribute 'chunked'".format(
                    klass.__name__
                )
            )

        # response body is (Body, BodyIter), but body.chunked is True:
        for klass in (Body, BodyIter):
            body = klass(chunked=True, state=0, content_length=17)
            bad = (200, 'OK', {'content-length': 17}, body)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                'response[3].chunked must be False; got True'
            )

        # response body is (ChunkedBody, ChunkedBodyIter), but 'content-length'
        # header is included:
        for klass in (ChunkedBody, ChunkedBodyIter):
            body = klass(state=0, chunked=True)
            bad = (200, 'OK', {'content-length': 17}, body)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                '{}: response body is {!r}, but {!r} header is included'.format(
                    'response[3]', klass, 'content-length'
                )
            )

        # response body is (ChunkedBody, ChunkedBodyIter), but
        # 'transfer-encoding' header is missing (should be filled in by server
        # before writing response):
        for klass in (ChunkedBody, ChunkedBodyIter):
            body = klass(chunked=True, state=0)
            good = (200, 'OK', {}, body)
            self.assertIsNone(
                rgi._validate_response(default_bodies, deepcopy(request), good)
            )

        # response body is (ChunkedBody, ChunkedBodyIter), but missing 'chunked'
        # attribute:
        for klass in (ChunkedBody, ChunkedBodyIter):
            body = klass(state=0)
            bad = (200, 'OK', {'transfer-encoding': 'chunked'}, body)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                "response[3]: {!r} object has no attribute 'chunked'".format(
                    klass.__name__
                )
            )

        # response body is (ChunkedBody, ChunkedBodyIter), but body.chunked is
        # False:
        for klass in (ChunkedBody, ChunkedBodyIter):
            body = klass(chunked=False, state=0)
            bad = (200, 'OK', {'transfer-encoding': 'chunked'}, body)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                'response[3].chunked must be True; got False'
            )

        # When response body is (ChunkedBody, ChunkedBodyIter), should not have
        # a 'content_length' attribute:
        for klass in (ChunkedBody, ChunkedBodyIter):
            body = klass(chunked=True, state=0, content_length=17)
            bad = (200, 'OK', {'transfer-encoding': 'chunked'}, body)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                "response[3]: {!r} must not have a 'content_length' attribute".format(
                    klass
                )
            )

        # response body is (Body, BodyIter, ChunkedBody, ChunkedBodyIter),
        # but missing 'state' attribute:
        bad_bodies = (
            Body(chunked=False, content_length=17),
            BodyIter(chunked=False, content_length=17),
            ChunkedBody(chunked=True),
            ChunkedBodyIter(chunked=True),
        )
        for body in bad_bodies:
            bad = (200, 'OK', {}, body)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                "response[3]: {!r} object has no attribute 'state'".format(
                    type(body).__name__
                )
            )

        # response body is (Body, BodyIter, ChunkedBody, ChunkedBodyIter),
        # but body.state != 0:
        bad_bodies = (
            Body(state=1, chunked=False, content_length=17),
            BodyIter(state=1, chunked=False, content_length=17),
            ChunkedBody(state=1, chunked=True),
            ChunkedBodyIter(state=1, chunked=True),
        )
        for body in bad_bodies:
            bad = (200, 'OK', {}, body)
            with self.assertRaises(ValueError) as cm:
                rgi._validate_response(default_bodies, deepcopy(request), bad)
            self.assertEqual(str(cm.exception),
                'response[3].state must equal 0; got 1'
            )


class TestValidator(TestCase):
    def test_init(self):
        # app not callable:
        class Bad:
            pass

        bad = Bad()
        with self.assertRaises(TypeError) as cm:
            rgi.Validator(bad)
        self.assertEqual(str(cm.exception),
            'app: not callable: {!r}'.format(bad)
        )

        # app.on_connect not callable:
        class Bad:
            def __init__(self):
                self.on_connect = random_identifier()

            def __call__(self, session, request):
                return (200, 'OK', {}, None)

        bad = Bad()
        with self.assertRaises(TypeError) as cm:
            rgi.Validator(bad)
        self.assertEqual(str(cm.exception),
            'app.on_connect: not callable: {!r}'.format(bad.on_connect)
        )

        # app is callable, no on_connect attribute:
        def good_app(session, request):
            return (200, 'OK', {}, None)

        inst = rgi.Validator(good_app)
        self.assertIs(inst.app, good_app)
        self.assertIsNone(inst._on_connect)

        # app is callable, on_connect attribute is None:
        class GoodApp:
            def __init__(self):
                self.on_connect = None

            def __call__(self, session, request):
                return (200, 'OK', {}, None)

        good_app = GoodApp()
        inst = rgi.Validator(good_app)
        self.assertIs(inst.app, good_app)
        self.assertIsNone(inst._on_connect)

        #app is callable, on_connect attribute also callable:
        class GoodApp:
            def __call__(self, session, request):
                return (200, 'OK', {}, None)

            def on_connect(self, sock, session):
                return True

        good_app = GoodApp()
        inst = rgi.Validator(good_app)
        self.assertIs(inst.app, good_app)
        self.assertEqual(inst._on_connect, good_app.on_connect)

    def test_repr(self):
        def my_app(session, request):
            return (200, 'OK', {'x-msg': 'hello, world'}, None)

        inst = rgi.Validator(my_app)
        self.assertEqual(repr(inst), 'Validator({!r})'.format(my_app))

        class Example(rgi.Validator):
            pass

        inst = Example(my_app)
        self.assertEqual(repr(inst), 'Example({!r})'.format(my_app))

    def test_call(self):
        # request.body.state != 2 after app() was called:
        def my_app(session, request, bodies):
            return (200, 'OK', {}, None)

        inst = rgi.Validator(my_app)
        session = build_session()
        request = build_request(
            body=Body(state=0, chunked=False, content_length=17),
            headers={'content-length': 17},
        )
        with self.assertRaises(ValueError) as cm:
            inst(session, request, default_bodies)
        self.assertEqual(str(cm.exception),
            'request.body.state must be 2 after app() was called; got 0'
        )

        # request.body.state == 2 after app() is called:
        def my_app(session, request, bodies):
            request.body.state = 2
            return (200, 'OK', {}, None)

        inst = rgi.Validator(my_app)
        session = build_session()
        request = build_request(
            body=Body(state=0, chunked=False, content_length=17),
            headers={'content-length': 17},
        )
        self.assertEqual(inst(session, request, default_bodies),
            (200, 'OK', {}, None)
        )

    def test_on_connect(self):
        class App:
            def __init__(self, allow):
                self.__allow = allow

            def __call__(self, session, request, bodies):
                raise Exception('should not be called')

            def on_connect(self, session, sock):
                if isinstance(self.__allow, Exception):
                    raise self.__allow
                return self.__allow

        # app.on_connect() doesn't return a bool:
        inst = rgi.Validator(App(1))
        session = build_session()
        with self.assertRaises(TypeError) as cm:
            inst.on_connect(session, None)
        self.assertEqual(str(cm.exception),
            "app.on_connect() must return a <class 'bool'>; got a <class 'int'>: 1"
        )
        inst = rgi.Validator(App('true'))
        session = build_session()
        with self.assertRaises(TypeError) as cm:
            inst.on_connect(session, None)
        self.assertEqual(str(cm.exception),
            "app.on_connect() must return a <class 'bool'>; got a <class 'str'>: 'true'"
        )

        # app.on_connect() raises an Exception:
        marker = random_identifier()
        inst = rgi.Validator(App(ValueError(marker)))
        session = build_session()
        with self.assertRaises(ValueError) as cm:
            inst.on_connect(session, None)
        self.assertEqual(str(cm.exception), marker)

        # app.on_connect() returns True:
        inst = rgi.Validator(App(True))
        session = build_session()
        self.assertIs(inst.on_connect(session, None), True)

        # app.on_connect() returns True, but session.requests != 0:
        inst = rgi.Validator(App(True))
        session = build_session(requests=1)
        with self.assertRaises(ValueError) as cm:
            inst.on_connect(session, None)
        self.assertEqual(str(cm.exception), 
            'session.requests must be 0 when app.on_connect() is called; got 1'
        )

        # app.on_connect() returns False:
        inst = rgi.Validator(App(False))
        session = build_session()
        self.assertIs(inst.on_connect(session, None), False)

        # app.on_connect() returns False, but session.requests != 0:
        inst = rgi.Validator(App(False))
        session = build_session(requests=3)
        with self.assertRaises(ValueError) as cm:
            inst.on_connect(session, None)
        self.assertEqual(str(cm.exception), 
            'session.requests must be 0 when app.on_connect() is called; got 3'
        )

        # app has no 'on_connect' attribute:
        def my_app(session, request):
            raise Exception('should not be called')

        inst = rgi.Validator(my_app)
        session = build_session()
        self.assertIs(inst.on_connect(session, None), True)

        # app.on_connect is None:
        my_app.on_connect = None
        inst = rgi.Validator(my_app)
        session = build_session()
        self.assertIs(inst.on_connect(session, None), True)

